import numpy as np
from context import arkouda as ak
from base_test import ArkoudaTest

N = 1_000_000

'''
The CompareTest class encapsulates unit tests that compare the results of analogous  
method invocations against Numpy ndarrays and Arkouda pdarrays to ensure equivalent 
results are generated.
'''
class CompareTest(ArkoudaTest):

    def test_compare_arange(self):
        # create np version
        nArange = np.arange(N)
        # create ak version
        aArange = ak.arange(N)
        self.assertEqual(nArange.all(), aArange.all())

    def test_compare_linspace(self):
        # create np version
        a = np.linspace(10, 20, N)
        # create ak version
        b = ak.linspace(10, 20, N)
        # print(a,b)
        self.assertTrue(np.allclose(a, b.to_ndarray()))

    def test_compare_ones(self):
        # create np version
        nArray = np.ones(N)
        # create ak version
        aArray = ak.ones(N)
        self.assertEqual(nArray.all(), aArray.to_ndarray().all())

    def test_compare_zeros(self):
        # create np version
        nArray = np.zeros(N)
        # create ak version
        aArray = ak.zeros(N)
        self.assertEqual(nArray.all(), aArray.to_ndarray().all())

    def test_compare_argsort(self):
        # create np version
        a = np.arange(N)
        a = a[::-1]
        iv = np.argsort(a)
        a = a[iv]
        # create ak version
        b = ak.arange(N)
        b = b[::-1]
        iv = ak.argsort(b)
        b = b[iv]

    def test_compare_sort(self):
        # create np version
        a = np.arange(N)
        a = a[::-1]
        a = np.sort(a)
        # create ak version
        b = ak.arange(N)
        b = b[::-1]
        b = ak.sort(b)
        self.assertEqual(a.all(), b.to_ndarray().all())

    def test_compare_get_slice(self):
        # create np version
        a = np.ones(N)
        a = a[::2]
        # create ak version
        b = ak.ones(N)
        b = b[::2]
        self.assertEqual(a.all(), b.to_ndarray().all())

    def test_compare_set_slice_value(self):
        # create np version
        a = np.ones(N)
        a[::2] = -1
        # create ak version
        b = ak.ones(N)
        b[::2] = -1
        self.assertEqual(a.all(), b.to_ndarray().all())

    def test_compare_set_slice(self):
        # create np version
        a = np.ones(N)
        a[::2] = a[::2] * -1
        # create ak version
        b = ak.ones(N)
        b[::2] = b[::2] * -1
        self.assertEqual(a.all(), b.to_ndarray().all())

    def test_compare_get_bool_iv(self):
        # create np version
        a = np.arange(N)
        a = a[a < N//2]
        # create ak version
        b = ak.arange(N)
        b = b[b < N//2]
        self.assertEqual(a.all(), b.to_ndarray().all())

    def test_compare_set_bool_iv_value(self):
        # create np version
        a = np.arange(N)
        a[a < N//2] = -1
        # create ak version
        b = ak.arange(N)
        b[b < N//2] = -1
        self.assertEqual(a.all(), b.to_ndarray().all())

    def check_set_bool_iv(self):
        # create np version
        a = np.arange(N)
        a[a < N//2] = a[:N//2] * -1
        # create ak version
        b = ak.arange(N)
        b[b < N//2] = b[:N//2] * -1
        self.assertEqual(a.all(), b.to_ndarray().all())
       
    def check_get_integer_iv(self):
        # create np version
        a = np.arange(N)
        iv = np.arange(N//2)
        a = a[iv]
        # create ak version
        b = ak.arange(N)
        iv = ak.arange(N//2)
        b = b[iv]
        self.assertEqual(a.all(), b.to_ndarray().all())

    def test_compare_set_integer_iv_val(self):
        # create np version
        a = np.arange(N)
        iv = np.arange(N//2)
        a[iv] = -1
        # create ak version
        b = ak.arange(N)
        iv = ak.arange(N//2)
        b[iv] = -1
        self.assertEqual(a.all(), b.to_ndarray().all())

    def test_compare_set_integer_iv(self):
        # create np version
        a = np.arange(N)
        iv = np.arange(N//2)
        a[iv] = iv*10
        # create ak version
        b = ak.arange(N)
        iv = ak.arange(N//2)
        b[iv] = iv*10
        self.assertEqual(a.all(), b.to_ndarray().all())

    def test_compare_get_integer_idx(self):
        # create np version
        a = np.arange(N)
        v1 = a[N//2]
        # create ak version
        b = ak.arange(N)
        v2 = b[N//2]
        self.assertEqual(v1, v2)
        self.assertEqual(a[-1], b[-1])

    def test_compare_set_integer_idx(self):
        # create np version
        a = np.arange(N)
        a[N//2] = -1
        a[-1] = -1
        v1 = a[N//2]
        # create ak version
        b = ak.arange(N)
        b[N//2] = -1
        b[-1] = -1
        v2 = b[N//2]
        self.assertEqual(v1, v2)
        self.assertEqual(a[-1], b[-1])



import unittest, os, sys
from context import arkouda as ak
from util.test.util import ServerInfo, get_arkouda_numlocales, start_arkouda_server, stop_arkouda_server, \
     run_client_live, set_server_info
import importlib

'''
ArkoudaTest defines the base Arkouda test logic for starting up the arkouda_server at the 
launch of a unittest TestCase and shutting down the arkouda_server at the completion of
the unittest TestCase.

Note: each Arkouda TestCase class extends ArkoudaTest and encompasses 1..n test methods, the 
names of which match the pattern 'test*' (e.g., ArkoudaTest.test_arkouda_server).
'''
class ArkoudaTest(unittest.TestCase):

    verbose = True if os.getenv('ARKOUDA_VERBOSE') == 'True' else False
    port = int(os.getenv('ARKOUDA_SERVER_PORT', 5555))
    server = os.getenv('ARKOUDA_SERVER_HOST', 'localhost')
    full_stack_mode = True if os.getenv('ARKOUDA_FULL_STACK_TEST') == 'True' else False
    timeout = int(os.getenv('ARKOUDA_CLIENT_TIMEOUT', 5))
    
    @classmethod
    def setUpClass(cls):
        '''
        If in full stack mode, configures and starts the arkouda_server process and sets the 
        ArkoudaTest.server class attribute, noop in client mode.
        
        :return: None
        :raise: EnvironmentError if 1..n required test libraries are missing, RuntimeError 
                if there is an error in configuring or starting arkouda_server
        '''
        if not importlib.util.find_spec('pytest') or not importlib.util.find_spec('pytest_env'):
            raise EnvironmentError('pytest and pytest-env must be installed')
        if ArkoudaTest.full_stack_mode:
            print('starting in full stack mode')
            try: 
                nl = get_arkouda_numlocales()
                ArkoudaTest.server, _, _ = start_arkouda_server(numlocales=nl, port=ArkoudaTest.port)
                print('Started arkouda_server in full stack test mode with host: {} port: {} locales: {}'.\
                      format(ArkoudaTest.server, ArkoudaTest.port, nl))
            except Exception as e:
                raise RuntimeError('in configuring or starting the arkouda_server: {}, check ' +
                         'environment and/or arkouda_server installation', e)
        else:
            print('in client stack test mode with host: {} port: {}'.format(ArkoudaTest.server, ArkoudaTest.port))
        

    def setUp(self):
        '''
        Connects an Arkouda client for each test case
        
        :return: None
        :raise: ConnectionError if exception is raised in connecting to arkouda_server
        '''
        try:
            ak.client.connect(server=ArkoudaTest.server, port=ArkoudaTest.port, 
                              timeout=ArkoudaTest.timeout)
        except Exception as e:
            raise ConnectionError(e)
    
    def tearDown(self):
        '''
        Disconnects the client connection for each test case
        :return: None
        :raise: ConnectionError if exception is raised in connecting to arkouda_server
        '''
        try:
            ak.client.disconnect()
        except Exception as e:
            raise ConnectionError(e)

    @classmethod
    def tearDownClass(cls):
        '''
        Shuts down the arkouda_server started in the setUpClass method if test is run
        in full stack mode, noop if in client stack mode.

        :return: None
        '''
        if ArkoudaTest.full_stack_mode:
            try:
                stop_arkouda_server()
            except Exception as e:
                raise RuntimeError(e)



#!/usr/bin/env python3                                                         

import time, argparse
import numpy as np
import arkouda as ak

OPS = ('cumsum', 'cumprod')
TYPES = ('int64', 'float64')

def time_ak_scan(N_per_locale, trials, dtype, random):
    print(">>> arkouda scan")
    cfg = ak.get_config()
    N = N_per_locale * cfg["numLocales"]
    print("numLocales = {}, N = {:,}".format(cfg["numLocales"], N))
    if random:
        if dtype == 'int64':
            a = ak.randint(0, 2**32, N)
        elif dtype == 'float64':
            a = ak.randint(0, 1, N, dtype=ak.float64)
    else:
        a = ak.arange(0, N, 1)
        if dtype == 'float64':
            a = 1.0 * a
     
    timings = {op: [] for op in OPS}
    final_values = {}
    for i in range(trials):
        for op in timings.keys():
            fxn = getattr(ak, op)
            start = time.time()
            r = fxn(a)
            end = time.time()
            timings[op].append(end - start)
            final_values[op] = r[r.size-1]
    tavg = {op: sum(t) / trials for op, t in timings.items()}

    for op, t in tavg.items():
        print("{}, final value = {}".format(op, final_values[op]))
        print("  {} Average time = {:.4f} sec".format(op, t))
        bytes_per_sec = (a.size * a.itemsize * 2) / t
        print("  {} Average rate = {:.2f} GiB/sec".format(op, bytes_per_sec/2**30))

def time_np_scan(N, trials, dtype, random):
    print(">>> numpy scan")
    print("N = {:,}".format(N))
    if random:
        if dtype == 'int64':
            a = np.random.randint(0, 2**32, N)
        elif dtype == 'float64':
            a = np.random.random(N)
    else:   
        a = np.arange(0, N, 1, dtype=dtype)
     
    timings = {op: [] for op in OPS}
    final_values = {}
    for i in range(trials):
        for op in timings.keys():
            fxn = getattr(np, op)
            start = time.time()
            r = fxn(a)
            end = time.time()
            timings[op].append(end - start)
            final_values[op] = r[r.size-1]
    tavg = {op: sum(t) / trials for op, t in timings.items()}

    for op, t in tavg.items():
        print("{}, final value = {}".format(op, final_values[op]))
        print("  {} Average time = {:.4f} sec".format(op, t))
        bytes_per_sec = (a.size * a.itemsize * 2) / t
        print("  {} Average rate = {:.2f} GiB/sec".format(op, bytes_per_sec/2**30))

def check_correctness(dtype, random):
    N = 10**4
    if random:
        if dtype == 'int64':
            a = np.random.randint(0, 2**32, N)
        elif dtype == 'float64':
            a = np.random.random(N)
    else:
        if dtype == 'int64':
            a = np.arange(0, N, 1, dtype=dtype)
        elif dtype == 'float64':
            a = np.arange(1, 1+1/N, (1/N)/N, dtype=dtype)

    for op in OPS:
        npa = a
        aka = ak.array(a)
        fxn = getattr(np, op)
        npr = fxn(npa)
        fxn = getattr(ak, op)
        akr = fxn(aka)
        assert np.allclose(npr, akr.to_ndarray())

def create_parser():
    parser = argparse.ArgumentParser(description="Measure the performance of scans (cumulative reductions) over arrays.")
    parser.add_argument('hostname', help='Hostname of arkouda server')
    parser.add_argument('port', type=int, help='Port of arkouda server')
    parser.add_argument('-n', '--size', type=int, default=10**8, help='Problem size: length of array')
    parser.add_argument('-t', '--trials', type=int, default=6, help='Number of times to run the benchmark')
    parser.add_argument('-d', '--dtype', default='int64', help='Dtype of array ({})'.format(', '.join(TYPES)))
    parser.add_argument('-r', '--randomize', default=False, action='store_true', help='Fill array with random values instead of range')
    parser.add_argument('--numpy', default=False, action='store_true', help='Run the same operation in NumPy to compare performance.')
    parser.add_argument('--correctness-only', default=False, action='store_true', help='Only check correctness, not performance.')
    return parser

if __name__ == "__main__":
    import sys
    parser = create_parser()
    args = parser.parse_args()
    if args.dtype not in TYPES:
        raise ValueError("Dtype must be {}, not {}".format('/'.join(TYPES), args.dtype))
    ak.verbose = False
    ak.connect(args.hostname, args.port)

    if args.correctness_only:
        for dtype in TYPES:
            check_correctness(dtype, args.randomize)
        sys.exit(0)
    
    print("array size = {:,}".format(args.size))
    print("number of trials = ", args.trials)
    time_ak_scan(args.size, args.trials, args.dtype, args.randomize)
    if args.numpy:
        time_np_scan(args.size, args.trials, args.dtype, args.randomize)
    sys.exit(0)



#!/usr/bin/env python3

import importlib
import numpy as np
import math
import gc
import sys

import arkouda as ak

ak.verbose = False
if len(sys.argv) > 1:
    ak.connect(server=sys.argv[1], port=sys.argv[2])
else:
    ak.connect()

a = ak.arange(0, 10, 1)
b = np.linspace(10, 20, 10)
c = ak.array(b)
d = a + c
e = d.to_ndarray()
    
a = ak.ones(10)
a[::2] = 0
print(a)

a = ak.ones(10)
b = ak.zeros(5)
a[1::2] = b
print(a)

a = ak.zeros(10) # float64
b = ak.arange(0,10,1) # int64
a[:] = b # cast b to float64
print(b,b.dtype)
print(a,a.dtype)

a = ak.randint(0,2,10,dtype=ak.bool)
a[1] = True
a[2] = True
print(a)

a = ak.ones(10,dtype=ak.int64)
iv = ak.arange(0,10,1)[::2]
a[iv] = 10
print(a)

a = ak.ones(10)
iv = ak.arange(0,10,1)[::2]
a[iv] = 10.0
print(a)

a = ak.ones(10,dtype=ak.bool)
iv = ak.arange(0,10,1)[::2]
a[iv] = False
print(a)

a = ak.ones(10,dtype=ak.bool)
iv = ak.arange(0,5,1)
b = ak.zeros(iv.size,dtype=ak.bool)
a[iv] = b
print(a)

a = ak.randint(10,20,10)
print(a)
iv = ak.randint(0,10,5)
print(iv)
b = ak.zeros(iv.size,dtype=ak.int64)
a[iv] = b
print(a)

ak.verbose = False
a = ak.randint(10,30,40)
vc = ak.value_counts(a)
print(vc[0].size,vc[0])
print(vc[1].size,vc[1])

ak.verbose = False

a = ak.arange(0,10,1)
b = a[a<5]
a = ak.linspace(0,9,10)
b = a[a<5]
print(b)

ak.verbose = True
ak.pdarrayIterThresh = 1000
a = ak.arange(0,10,1)
print(list(a))

ak.verbose = False
a = ak.randint(10,30,40)
u = ak.unique(a)
h = ak.histogram(a,bins=20)
print(a)
print(h.size,h)
print(u.size,u)

ak.verbose = False
a = ak.randint(10,30,50)
h = ak.histogram(a,bins=20)
print(a)
print(h)

ak.verbose = False
a = ak.randint(0,2,50,dtype=ak.bool)
print(a)
print(a.sum())

ak.verbose = False
a = ak.linspace(101,102,100)
h = ak.histogram(a,bins=50)
print(h)

ak.verbose = False
a = ak.arange(0,100,1)
h = ak.histogram(a,bins=10)
print(h)

ak.verbose = False
a = ak.arange(0,99,1)
b = a[::10] # take every tenth one
b = b[::-1] # reverse b
print(a)
print(b)
c = ak.in1d(a,b) # put out truth vector
print(c)
print(a[c]) # compress out false values

ak.verbose = False
a = np.ones(10,dtype=np.bool)
b = np.arange(0,10,1)

np.sum(a),np.cumsum(a),np.sum(b<4),b[b<4],b<5

ak.verbose = False
# currently... ak pdarray to np array
a = ak.linspace(0,9,10)
b = np.array(list(a))
print(a,a.dtype,b,b.dtype)

a = ak.arange(0,10,1)
b = np.array(list(a))
print(a,a.dtype,b,b.dtype)

a = ak.ones(10,dtype=ak.bool)
b = np.array(list(a))
print(a,a.dtype,b,b.dtype)

ak.verbose = False

b = np.linspace(1,10,10)
a = np.arange(1,11,1)
print(b/a)

ak.verbose = False

#a = np.ones(10000,dtype=np.int64)
a = np.linspace(0,99,100)
#a = np.arange(0,100,1)
print(a)

ak.verbose = False
print(a.__repr__())
print(a.__str__())

ak.verbose = False
print(a)
print(type(a), a.dtype, a.size, a.ndim, a.shape, a.itemsize)

ak.verbose = False
a = ak.arange(0,100,1)

ak.verbose = False
print(a)

ak.verbose = False

b = ak.linspace(0,99,100)
print(b.__repr__())

ak.verbose = False
b = ak.linspace(0,9,10)
a = ak.arange(0,10,1)
print(a.name, a.size, a.dtype, a)
print(b.name, b.size, b.dtype, b)
print(ak.info(a+b))


ak.verbose = False

c = ak.arange(0,10,1)
print(ak.info(c))
print(c.name, c.dtype, c.size, c.ndim, c.shape, c.itemsize)
print(c)

ak.verbose = False

print(5+c + 5)

c = np.array([10, 11, 12, 13, 14, 15, 16, 17, 18, 19])
c = np.array([10, 11, 12, 13, 14, 15, 16, 17, 18, 19.1])
print(c.__repr__(), c.dtype.__str__(), c.dtype.__repr__())

ak.verbose = False
a = np.ones(9)
b = np.arange(1,10,1)
print(a.dtype,b.dtype)
c = ak.ones(9)
d = ak.arange(1,10,1)
print(c.dtype,d.dtype)
y = a/b
z = c/d
print("truediv  \nnp out:",y,"\nak out:",z)
print(y[5],z[5],y[5] ==z[5])
y = a//b
z = c//d
print("floordiv \nnp out:",y,"\nak out:",z)
print(y[5],z[5],y[5] ==z[5])

ak.verbose = False

c = ak.arange(1,10,1)
c //= c
print(c)
c += c
print(c)
c *= c
print(c)

ak.verbose = False

a = np.ones(9,dtype=np.int64)
b = np.ones_like(a)
print(b)

ak.verbose = False

a = ak.ones(9,dtype=ak.int64)
b = ak.ones_like(a)
print(b)

ak.verbose = False

a = ak.arange(0,10,1)
b = np.arange(0,10,1)

print(a[5] == b[5])

ak.verbose = False

a = ak.arange(0,10,1)
b = np.arange(0,10,1)

a[5] = 10.2
print(a[5])

ak.verbose = False

a = ak.arange(0,10,1)
b = np.arange(0,10,1)
#print((a[:]),b[:])
#print(a[1:-1:2],b[1:-1:2])
#print(a[0:10:2],b[0:10:2])
print(a[4:20:-1],b[4:20:-1])
print(a[:1:-1],b[:1:-1])

ak.verbose = False
d = ak.arange(1,10,1)
#d.type.__class__,d.name,d.isnative,np.int64.__class__,bool
ak.info(d)
#dir(d)

ak.verbose = False

a = ak.ones(10,dtype=ak.bool)
print(a[1])

ak.verbose = False

a = ak.zeros(10,dtype=ak.bool)
print(a[1])

ak.verbose = False
a = ak.ones(10,dtype=ak.bool)
a[4] = False
a[1] = False
print(a)
print(a[::2])
print(a[1])
a = ak.ones(10,dtype=ak.int64)
a[4] = False
a[1] = False
print(a)
print(a[::2])
print(a[1])
a = ak.ones(10)
a[4] = False
a[1] = False
print(a)
print(a[::2])
print(a[1])

ak.verbose = False

a = ak.arange(0,10,1)
b = list(a)
print(b)

a = a<5
b = list(a)
print(b)

ak.verbose = False
a = ak.linspace(1,10,10)
print(ak.abs(a))
print(ak.log(a))
print(ak.exp(a))
a.fill(math.e)
print(ak.log(a))

type(bool),type(np.bool),type(ak.bool),type(True)

ak.verbose = False
a = ak.linspace(0,9,10)
print(a,ak.any(a),ak.all(a),ak.all(ak.ones(10,dtype=ak.float64)))
b = a<5
print(b,ak.any(b),ak.all(b),ak.all(ak.ones(10,dtype=ak.bool)))
c = ak.arange(0,10,1)
print(c,ak.any(c),ak.all(c),ak.all(ak.ones(10,dtype=ak.int64)))
print(a.any(),a.all(),b.any(),b.all())

ak.verbose = False
a = ak.linspace(0,9,10)
ak.sum(a)
b = np.linspace(0,9,10)
print(ak.sum(a) == np.sum(b),ak.sum(a),np.sum(b),a.sum(),b.sum())

ak.verbose = False
a = ak.linspace(1,10,10)
b = np.linspace(1,10,10)
print(ak.prod(a) == np.prod(b),ak.prod(a),np.prod(b),a.prod(),b.prod())

ak.verbose = False

a = np.arange(0,20,1)
b = a<10
print(b,np.sum(b),b.sum(),np.prod(b),b.prod(),np.cumsum(b),np.cumprod(b))
print()
b = a<5
print(b,np.sum(b),b.sum(),np.prod(b),b.prod(),np.cumsum(b),np.cumprod(b))
print()
a = ak.arange(0,20,1)
b = a<10
print(b,ak.sum(b),b.sum(),ak.prod(b),b.prod(),ak.cumsum(b),ak.cumprod(b))
b = a<5
print(b,ak.sum(b),b.sum(),ak.prod(b),b.prod(),ak.cumsum(b),ak.cumprod(b))

ak.verbose = False
a = ak.arange(0,10,1)
iv = a[::-1]
print(a,iv,a[iv])

ak.verbose = False
a = ak.arange(0,10,1)
iv = a[::-1]
print(a,iv,a[iv])

ak.verbose = False
a = ak.linspace(0,9,10)
iv = ak.arange(0,10,1)
iv = iv[::-1]
print(a,iv,a[iv])

ak.verbose = False
a = np.arange(0,10,1)
iv = a[::-1]
print(a,iv,a[iv])

ak.verbose = False
a = ak.arange(0,10,1)
b = a<20
print(a,b,a[b])

ak.verbose = False
a = ak.arange(0,10,1)
b = a<5
print(a,b,a[b])

ak.verbose = False
a = ak.arange(0,10,1)
b = a<0
print(a,b,a[b])

ak.verbose = False
a = ak.linspace(0,9,10)
b = a<5
print(a,b,a[b])

ak.verbose = False
N = 2**23 # 2**23 * 8 == 64MiB
A = ak.linspace(0,N-1,N)
B = ak.linspace(0,N-1,N)

C = A+B
print(ak.info(C),C)

# turn off verbose messages from arkouda package
ak.verbose = False
# set pdarrayIterThresh to 0 to only print the first 3 and last 3 of pdarray
ak.pdarrayIterThresh = 0
a = ak.linspace(0,9,10)
b = a<5
print(a)
print(b)
print(a[b])
print(a)

a = np.linspace(0,9,10)
b = a<5
print(a)
print(b)
print(a[b])
print(a)

ak.verbose = False
ak.pdarrayIterThresh = 0

a = ak.ones(10,ak.int64)
b = a | 0xff
print(a, b, a^b, b>>a, b<<1|1, 0xf & b, 0xaa ^ b, b ^ 0xff)
print(-a,~(~a))

a = ak.ones(10,dtype=ak.int64)
b = ~ak.zeros(10,dtype=ak.int64)
print(~a, -b)

a = np.ones(10,dtype=np.int64)
b = ~np.zeros(10,dtype=np.int64)
print(~a, -b)


ak.shutdown()




#!/usr/bin/env python3

import sys, time, argparse
import arkouda as ak

if __name__ == '__main__':
    ak.verbose = False
    parser = argparse.ArgumentParser()
    parser.add_argument('--server', default=None)
    parser.add_argument('--port', default=None)
    parser.add_argument('dsetName')
    parser.add_argument('filenames', nargs='+')

    args = parser.parse_args()
    if args.server is not None:
        if args.port is not None:
            ak.connect(server=args.server, port=args.port)
        else:
            ak.connect(server=args.server)
    else:
        if args.port is not None:
            ak.connect(port=args.port)
        else:
            ak.connect()
    print("Reading files...")
    start = time.time()
    a = ak.read_hdf(args.dsetName, args.filenames)
    end = time.time()
    t = end - start
    print(a)
    print(f'{t:.2f} seconds ({8*a.size/t:.2e} bytes/sec)')
    print("Testing bad filename...")
    badfilename = args.filenames[0] + '-should-not-exist-5473219431'
    try:
        ak.read_hdf(args.dsetName, args.filenames + [badfilename])
    except RuntimeError as e:
        print(e)
    print("Testing bad dsetName...")
    try:
        ak.read_hdf(args.dsetName+'-not-a-dset', args.filenames)
    except RuntimeError as e:
        print(e)
        
    ak.shutdown()
    sys.exit()



#!/usr/bin/env python3
"""
This is a driver script to automatically run the Arkouda benchmarks in this
directory and optionally graph the results. Graphing requires that $CHPL_HOME
points to a valid Chapel directory. This will start and stop the Arkouda server
automatically.
"""

import argparse
import logging
import os
import subprocess
import sys

benchmark_dir = os.path.dirname(__file__)
util_dir = os.path.join(benchmark_dir, '..', 'util', 'test')
sys.path.insert(0, os.path.abspath(util_dir))
from util import *

logging.basicConfig(level=logging.INFO)

BENCHMARKS = ['stream', 'argsort', 'coargsort', 'gather', 'scatter', 'reduce', 'scan', 'noop', 'setops']

def get_chpl_util_dir():
    """ Get the Chapel directory that contains graph generation utilities. """
    CHPL_HOME = os.getenv('CHPL_HOME')
    if not CHPL_HOME:
        logging.error('$CHPL_HOME not set')
        sys.exit(1)
    chpl_util_dir = os.path.join(CHPL_HOME, 'util', 'test')
    if not os.path.isdir(chpl_util_dir):
        logging.error('{} does not exist'.format(chpl_util_dir))
        sys.exit(1)
    return chpl_util_dir

def add_to_dat(benchmark, output, dat_dir, graph_infra):
    """
    Run computePerfStats to take output from a benchmark and create/append to a
    .dat file that contains performance keys. The performance keys come from
    `graph_infra/<benchmark>.perfkeys` if it exists, otherwise a default
    `graph_infra/perfkeys` is used.
    """
    computePerfStats = os.path.join(get_chpl_util_dir(), 'computePerfStats')

    perfkeys = os.path.join(graph_infra, '{}.perfkeys'.format(benchmark))
    if not os.path.exists(perfkeys):
        perfkeys = os.path.join(graph_infra, 'perfkeys')

    benchmark_out = '{}.exec.out.tmp'.format(benchmark)
    with open (benchmark_out, 'w') as f:
        f.write(output)
    subprocess.check_output([computePerfStats, benchmark, dat_dir, perfkeys, benchmark_out])
    os.remove(benchmark_out)

def generate_graphs(args):

    """
    Generate graphs using the existing .dat files and graph infrastructure.
    """
    genGraphs = os.path.join(get_chpl_util_dir(), 'genGraphs')
    cmd = [genGraphs,
           '--perfdir', args.dat_dir,
           '--outdir', args.graph_dir,
           '--graphlist', os.path.join(args.graph_infra, 'GRAPHLIST'),
           '--testdir', args.graph_infra,
           '--alttitle', 'Arkouda Performance Graphs']

    if args.platform_name:
        cmd += ['--name', args.platform_name]
    if args.configs:
        cmd += ['--configs', args.configs]
    if args.start_date:
        cmd += ['--startdate', args.start_date]
    if args.annotations:
        cmd += ['--annotate', args.annotations]


    subprocess.check_output(cmd)

def create_parser():
    parser = argparse.ArgumentParser(description=__doc__)

    # TODO support alias for a larger default N
    #parser.add_argument('--large', default=False, action='store_true', help='Run a larger problem size')

    parser.add_argument('-nl', '--num-locales', default=get_arkouda_numlocales(), help='Number of locales to use for the server')
    parser.add_argument('--numtrials', default=1, type=int, help='Number of trials to run')
    parser.add_argument('benchmarks', nargs='*', help='Basename of benchmarks to run with extension stripped')
    parser.add_argument('--gen-graphs', default=False, action='store_true', help='Generate graphs, requires $CHPL_HOME')
    parser.add_argument('--dat-dir', default=os.path.join(benchmark_dir, 'datdir'), help='Directory with .dat files stored')
    parser.add_argument('--graph-dir', help='Directory to place generated graphs')
    parser.add_argument('--graph-infra', default=os.path.join(benchmark_dir, 'graph_infra'), help='Directory containing graph infrastructure')
    parser.add_argument('--platform-name', default='', help='Test platform name')
    parser.add_argument('--description', default='', help='Description of this configuration')
    parser.add_argument('--annotations', default='', help='File containing annotations')
    parser.add_argument('--configs', help='comma seperate list of configurations')
    parser.add_argument('--start-date', help='graph start date')
    return parser

def main():
    parser = create_parser()
    args, client_args = parser.parse_known_args()
    args.graph_dir = args.graph_dir or os.path.join(args.dat_dir, 'html')
    config_dat_dir = os.path.join(args.dat_dir, args.description)

    if args.gen_graphs:
        os.makedirs(config_dat_dir, exist_ok=True)

    start_arkouda_server(args.num_locales)

    args.benchmarks = args.benchmarks or BENCHMARKS
    for benchmark in args.benchmarks:
        for trial in range(args.numtrials):
            benchmark_py = os.path.join(benchmark_dir, '{}.py'.format(benchmark))
            out = run_client(benchmark_py, client_args)
            if args.gen_graphs:
                add_to_dat(benchmark, out, config_dat_dir, args.graph_infra)
            print(out)

    stop_arkouda_server()

    if args.gen_graphs:
        comp_file = os.getenv('ARKOUDA_PRINT_PASSES_FILE', '')
        if os.path.isfile(comp_file):
            with open (comp_file, 'r') as f:
                out = f.read()
            add_to_dat('comp-time', out, config_dat_dir, args.graph_infra)
        generate_graphs(args)

if __name__ == '__main__':
    main()



from context import arkouda as ak
import numpy as np
import pandas as pd
from context import arkouda as ak
from base_test import ArkoudaTest

if __name__ == "__main__":
    import argparse, sys
    parser = argparse.ArgumentParser(description="Runs and times reductions over arrays in both arkouda and numpy.")
    parser.add_argument('hostname', help='Hostname of arkouda server')
    parser.add_argument('port', type=int, help='Port of arkouda server')
    parser.add_argument('-n', '--size', type=int, default=10**6, help='Problem size: length of array to argsort')

    args = parser.parse_args()
    ak.verbose = False
    ak.connect(args.hostname, args.port)
    print("size = ",args.size)
    SIZE = args.size
    a = ak.randint(0, 2*SIZE, SIZE)
    b = ak.randint(0, 2*SIZE, SIZE)

    
    set_union = ak.union1d(a,b)
    print("union1d = ", set_union.size,set_union)
    # elements in a or elements in b (or in both a and b)
    passed = ak.all(ak.in1d(set_union,a) | ak.in1d(set_union,b))
    print("union1d passed test: ",passed)
    
    set_intersection = ak.intersect1d(a,b)
    print("intersect1d = ", set_intersection.size,set_intersection)
    # elements in a and elements in b (elements in both a and b)
    passed = ak.all(ak.in1d(set_intersection,a) & ak.in1d(set_intersection,b))
    print("intersect1d passed test: ",passed)
    
    set_difference = ak.setdiff1d(a,b)
    print("setdiff1d = ", set_difference.size,set_difference)
    # elements in a and not in b
    passed = ak.all(ak.in1d(set_difference,a) & ak.in1d(set_difference,b,invert=True))
    print("setdiff1d passed test: ",passed)
    
    set_xor = ak.setxor1d(a,b)
    print("setxor1d = ", set_xor.size,set_xor)
    # elements NOT in the intersection of a and b
    passed = ak.all(ak.in1d(set_xor, set_intersection, invert=True))
    print("setxor1d passed test: ",passed)

    ak.disconnect()

class SetOpsTest(ArkoudaTest):

    def setUp(self):
        ArkoudaTest.setUp(self)
        SIZE = 5
        self.a = ak.randint(0, 2*SIZE, SIZE)
        self.b = ak.randint(0, 2*SIZE, SIZE)     
  
    def testOneDimensionalSetUnion(self): 
        set_union = ak.union1d(self.a,self.b)
        # elements in a or elements in b (or in both a and b)
        self.assertTrue(ak.all(ak.in1d(set_union,self.a) \
                               | ak.in1d(set_union,self.b)))

    def testOneDimensionalSetIntersection(self):
        set_intersection = ak.intersect1d(self.a,self.b)
        # elements in a and elements in b (elements in both a and b)
        self.assertTrue(ak.all(ak.in1d(set_intersection,self.a) & \
                               ak.in1d(set_intersection,self.b)))

    def testOneDimensionalSetDifference(self):
        set_difference = ak.setdiff1d(self.a,self.b)
        # elements in a and not in b
        self.assertTrue((ak.in1d(set_difference,self.a).all() & \
                         ak.in1d(set_difference,self.b,invert=True).all()))

    def testOneDimensionalSetXor(self):
        set_xor = ak.setxor1d(self.a,self.b)
        set_intersection = ak.intersect1d(self.a,self.b)
        # elements NOT in the intersection of a and b
        self.assertTrue(ak.all(ak.in1d(set_xor, set_intersection, invert=True)))


#!/usr/bin/env python3                                                         

import importlib
import numpy as np
import math
import gc
import sys

import arkouda as ak

ak.verbose = False
if len(sys.argv) > 1:
    ak.connect(server=sys.argv[1], port=sys.argv[2])
else:
    ak.connect()

N = 10**9
a = ak.ones(N,dtype='int64')
b = ak.ones(N,dtype='int64')
print(a,b)

c = a+b
d = a-b
print(c)
print(d)

ak.shutdown()




#!/usr/bin/env python3

import time, argparse
import numpy as np
import arkouda as ak
import arkouda.client

def time_ak_noop(trial_time):
    print(">>> arkouda noop")
    start = time.time()
    trials = 0
    while time.time() - start < trial_time:
        trials += 1
        arkouda.client._no_op()
    end = time.time()

    timing = end - start
    tavg = timing / trials

    print("Average time = {:.6f} sec".format(tavg))
    print("Average rate = {:.2f} ops/sec".format(trials/timing))


def time_np_noop(trial_time):
    print(">>> numpy noop")
    start = time.time()
    trials = 0
    while time.time() - start < trial_time:
        trials += 1
        np.get_include() # closet I could find to a noop
    end = time.time()

    timing = end - start
    tavg = timing / trials

    print("Average time = {:.6f} sec".format(tavg))
    print("Average rate = {:.2f} ops/sec".format(trials/timing))

def check_correctness():
    assert arkouda.client._no_op() == "noop"

def create_parser():
    parser = argparse.ArgumentParser(description="Run a noop benchmark")
    parser.add_argument('hostname', help='Hostname of arkouda server')
    parser.add_argument('port', type=int, help='Port of arkouda server')
    parser.add_argument('-n', '--size', type=int, default=1, help='Problem size (unused)')
    parser.add_argument('-t', '--trials', '--trials-time', type=int, default=1, help='Amount of time to run the benchmark')
    parser.add_argument('-d', '--dtype', default='int64', help='Dtype of arrays (unused)')
    parser.add_argument('--numpy', default=False, action='store_true', help='Run the same operation in NumPy to compare performance.')
    parser.add_argument('--correctness-only', default=False, action='store_true', help='Only check correctness, not performance.')
    return parser

if __name__ == "__main__":
    import sys
    parser = create_parser()
    args = parser.parse_args()

    ak.verbose = False
    ak.connect(args.hostname, args.port)

    if args.correctness_only:
        check_correctness()
        sys.exit(0)

    print("number of trials = ", args.trials)
    time_ak_noop(args.trials)
    if args.numpy:
        time_np_noop(args.size)
    sys.exit(0)



#!/usr/bin/env python3                                                         

import importlib
import numpy as np
import math
import gc
import sys

from context import arkouda as ak

print(">>> Sanity checks on the arkouda_server")

ak.verbose = False
if len(sys.argv) > 1:
    ak.connect(server=sys.argv[1], port=sys.argv[2])
else:
    ak.connect()


N = 1_000_000


errors = False
def pass_fail(f):
    global errors
    errors = errors or not f
    return ("Passed" if f else "Failed")

def check_bool(N):
    a = ak.arange(N)
    b = ak.ones(N)
    try:
        c = a and b
    except ValueError:
        correct = True
    except:
        correct = False
    d = ak.array([1])
    correct = correct and (d and 5)
    return pass_fail(correct)

print("check boolean :", check_bool(N))

def check_arange(N):
    # create np version
    a = np.arange(N)
    # create ak version
    b = ak.arange(N)
    # print(a,b)
    c = a == b.to_ndarray()
    # print(type(c),c)
    return pass_fail(c.all())
    
print("check arange :", check_arange(N))

def check_linspace(N):
    # create np version
    a = np.linspace(10, 20, N)
    # create ak version
    b = ak.linspace(10, 20, N)
    # print(a,b)
    f = np.allclose(a, b.to_ndarray())
    return pass_fail(f)

print("check linspace :", check_linspace(N))

def check_ones(N):
    # create np version
    a = np.ones(N)
    # create ak version
    b = ak.ones(N)
    # print(a,b)
    c = a == b.to_ndarray()
    # print(type(c),c)
    return pass_fail(c.all())

print("check ones :", check_ones(N))

def check_zeros(N):
    # create np version
    a = np.zeros(N)
    # create ak version
    b = ak.zeros(N)
    # print(a,b)
    c = a == b.to_ndarray()
    # print(type(c),c)
    return pass_fail(c.all())

print("check zeros :", check_zeros(N))

def check_argsort(N):
    # create np version
    a = np.arange(N)
    a = a[::-1]
    iv = np.argsort(a)
    a = a[iv]
    # create ak version
    b = ak.arange(N)
    b = b[::-1]
    iv = ak.argsort(b)
    b = b[iv]
    # print(a,b)
    c = a == b.to_ndarray()
    # print(type(c),c)
    return pass_fail(c.all())

print("check argsort :", check_argsort(N))

def check_coargsort(N):
    # create np version
    a = np.arange(N)
    a = a[::-1]
    iv = np.lexsort([a, a])
    a = a[iv]
    # create ak version
    b = ak.arange(N)
    b = b[::-1]
    iv = ak.coargsort([b, b])
    b = b[iv]
    # print(a,b)
    c = a == b.to_ndarray()
    # print(type(c),c)
    return pass_fail(c.all())

print("check coargsort :", check_coargsort(N))

def check_sort(N):
    # create np version
    a = np.arange(N)
    a = a[::-1]
    a = np.sort(a)
    # create ak version
    b = ak.arange(N)
    b = b[::-1]
    b = ak.sort(b)
    # print(a,b)
    c = a == b.to_ndarray()
    # print(type(c),c)
    return pass_fail(c.all())

print("check sort :", check_sort(N))

def check_get_slice(N):
    # create np version
    a = np.ones(N)
    a = a[::2]
    # create ak version
    b = ak.ones(N)
    b = b[::2]
    # print(a,b)
    c = a == b.to_ndarray()
    return pass_fail(c.all())

print("check get slice [::2] :", check_get_slice(N))

def check_set_slice_value(N):
    # create np version
    a = np.ones(N)
    a[::2] = -1
    # create ak version
    b = ak.ones(N)
    b[::2] = -1
    # print(a,b)
    c = a == b.to_ndarray()
    return pass_fail(c.all())

print("check set slice [::2] = value:", check_set_slice_value(N))

def check_set_slice(N):
    # create np version
    a = np.ones(N)
    a[::2] = a[::2] * -1
    # create ak version
    b = ak.ones(N)
    b[::2] = b[::2] * -1
    # print(a,b)
    c = a == b.to_ndarray()
    return pass_fail(c.all())

print("check set slice [::2] = pda:", check_set_slice(N))

def check_get_bool_iv(N):
    # create np version
    a = np.arange(N)
    a = a[a < N//2]
    # create ak version
    b = ak.arange(N)
    b = b[b < N//2]
    # print(a,b)
    c = a == b.to_ndarray()
    # print(type(c),c)
    return pass_fail(c.all())

print("check (compressing) get bool iv :", check_get_bool_iv(N))

def check_set_bool_iv_value(N):
    # create np version
    a = np.arange(N)
    a[a < N//2] = -1
    # create ak version
    b = ak.arange(N)
    b[b < N//2] = -1
    # print(a,b)
    c = a == b.to_ndarray()
    # print(type(c),c)
    return pass_fail(c.all())

print("check (expanding) set bool iv = value:", check_set_bool_iv_value(N))

def check_set_bool_iv(N):
    # create np version
    a = np.arange(N)
    a[a < N//2] = a[:N//2] * -1
    # create ak version
    b = ak.arange(N)
    b[b < N//2] = b[:N//2] * -1
    # print(a,b)
    c = a == b.to_ndarray()
    # print(type(c),c)
    return pass_fail(c.all())

print("check (expanding) set bool iv = pda:", check_set_bool_iv(N))

def check_get_integer_iv(N):
    # create np version
    a = np.arange(N)
    iv = np.arange(N//2)
    a = a[iv]
    # create ak version
    b = ak.arange(N)
    iv = ak.arange(N//2)
    b = b[iv]
    # print(a,b)
    c = a == b.to_ndarray()
    # print(type(c),c)
    return pass_fail(c.all())

print("check (gather) get integer iv:", check_get_integer_iv(N))

def check_set_integer_iv_value(N):
    # create np version
    a = np.arange(N)
    iv = np.arange(N//2)
    a[iv] = -1
    # create ak version
    b = ak.arange(N)
    iv = ak.arange(N//2)
    b[iv] = -1
    # print(a,b)
    c = a == b.to_ndarray()
    # print(type(c),c)
    return pass_fail(c.all())

print("check (scatter) set integer iv = value:", check_set_integer_iv_value(N))

def check_set_integer_iv(N):
    # create np version
    a = np.arange(N)
    iv = np.arange(N//2)
    a[iv] = iv*10
    # create ak version
    b = ak.arange(N)
    iv = ak.arange(N//2)
    b[iv] = iv*10
    # print(a,b)
    c = a == b.to_ndarray()
    # print(type(c),c)
    return pass_fail(c.all())

print("check (scatter) set integer iv = pda:", check_set_integer_iv(N))

def check_get_integer_idx(N):
    # create np version
    a = np.arange(N)
    v1 = a[N//2]
    # create ak version
    b = ak.arange(N)
    v2 = b[N//2]
    return pass_fail(v1 == v2) and pass_fail(a[-1] == b[-1])

print("check get integer idx :", check_get_integer_idx(N))

def check_set_integer_idx(N):
    # create np version
    a = np.arange(N)
    a[N//2] = -1
    a[-1] = -1
    v1 = a[N//2]
    # create ak version
    b = ak.arange(N)
    b[N//2] = -1
    b[-1] = -1
    v2 = b[N//2]
    return pass_fail(v1 == v2) and pass_fail(a[-1] == b[-1])

print("check set integer idx = value:", check_set_integer_idx(N))

ak.disconnect()
sys.exit(errors)



#!/usr/bin/env python3

import arkouda as ak
import sys, os

saveone = '/tmp/ak_save.hdf'
saveall = '/tmp/ak_save_all.hdf'

if len(sys.argv) < 4:
    print("Usage: {} <hostname> <port> <HDF5_filenames>".format(sys.argv[0]))
    sys.exit()
ak.connect(sys.argv[1], sys.argv[2])
onefile = sys.argv[3]
print(ak.ls_hdf(onefile))
allfiles = sys.argv[3:]
print(f"srcIP = ak.read_hdf('srcIP', {onefile})")
srcIP = ak.read_hdf('srcIP', onefile)
print(f"srcIP.save({saveone}, 'srcIP')")
srcIP.save(saveone, 'srcIP')
print(f"srcIP2 = ak.load({saveone}, 'srcIP')")
srcIP2 = ak.load(saveone, 'srcIP')
assert (srcIP == srcIP2).all()
del srcIP
del srcIP2
print(f"df = ak.read_all(['srcPort', 'proto', 'packets'], {allfiles})")
df = ak.read_all(['srcPort', 'proto', 'packets'], allfiles)
print(f"ak.save_all(df, {saveall})")
ak.save_all(df, saveall)
print(f"newdf = ak.load_all({saveall})")
newdf = ak.load_all(saveall)
print(newdf)
os.system('rm -rf /tmp/ak_save*')
ak.disconnect()



from arkouda.client import generic_msg, verbose
from arkouda.pdarrayclass import pdarray, create_pdarray
from arkouda.sorting import argsort, coargsort, local_argsort
from arkouda.strings import Strings
from arkouda.pdarraycreation import array, zeros, arange
from arkouda.pdarraysetops import concatenate
from arkouda.numeric import cumsum

__all__ = ["GroupBy"]

class GroupBy:
    """
    Group an array or list of arrays by value, usually in preparation 
    for aggregating the within-group values of another array.

    Parameters
    ----------
    keys : (list of) pdarray, int64 or Strings
        The array to group by value, or if list, the column arrays to group by row
    assume_sorted : bool
        If True, assume keys is already sorted (Default: False)

    Attributes
    ----------
    nkeys : int
        The number of key arrays (columns)
    size : int
        The length of the array(s), i.e. number of rows
    permutation : pdarray
        The permutation that sorts the keys array(s) by value (row)
    unique_keys : (list of) pdarray or Strings
        The unique values of the keys array(s), in grouped order
    segments : pdarray
        The start index of each group in the grouped array(s)
    unique_key_indices : pdarray
        The first index in the raw (ungrouped) keys array(s) where each 
        unique value (row) occurs

    Notes
    -----
    Only accepts pdarrays of int64 dtype or Strings.

    """
    Reductions = frozenset(['sum', 'prod', 'mean',
                            'min', 'max', 'argmin', 'argmax',
                            'nunique', 'any', 'all'])
    def __init__(self, keys, assume_sorted=False, hash_strings=True):
        self.assume_sorted = assume_sorted
        self.hash_strings = hash_strings
        self.per_locale = False
        self.keys = keys
        if isinstance(keys, pdarray):
            self.nkeys = 1
            self.size = keys.size
            if assume_sorted:
                self.permutation = arange(self.size)
            elif self.per_locale:
                self.permutation = local_argsort(keys)
            else:
                self.permutation = argsort(keys)
        # for Strings or Categorical
        elif hasattr(keys, "group"):
            self.nkeys = 1
            self.size = keys.size
            if assume_sorted:
                self.permutation = arange(self.size)
            elif self.per_locale:
                raise ValueError("per-locale groupby not supported on Strings or Categorical")
            else:
                self.permutation = keys.group()
        else:
            self.nkeys = len(keys)
            self.size = keys[0].size
            for k in keys:
                if k.size != self.size:
                    raise ValueError("Key arrays must all be same size")
            if assume_sorted:
                self.permutation = arange(self.size)
            else:
                self.permutation = coargsort(keys)
            
        # self.permuted_keys = self.keys[self.permutation]
        self.find_segments()
            
    def find_segments(self):
        if self.per_locale:
            cmd = "findLocalSegments"
        else:
            cmd = "findSegments"
        if self.nkeys == 1:
            # for Categorical
            if hasattr(self.keys, 'segments') and self.keys.segments is not None:
                self.unique_keys = self.keys.categories
                self.segments = self.keys.segments
                return
            else:
                mykeys = [self.keys]            
        else:
            mykeys = self.keys
        keynames = []
        keytypes = []
        effectiveKeys = self.nkeys
        for k in mykeys:
            if isinstance(k, Strings):
                if self.hash_strings:
                    h1, h2 = k.hash()
                    keynames.extend([h1.name, h2.name])
                    keytypes.extend([h1.objtype, h2.objtype])
                    effectiveKeys += 1
                else:
                    keynames.append('{}+{}'.format(k.offsets.name, k.bytes.name))
                    keytypes.append(k.objtype)
            # for Categorical
            elif hasattr(k, 'codes'):
                keynames.append(k.codes.name)
                keytypes.append(k.codes.objtype)
            elif isinstance(k, pdarray):
                keynames.append(k.name)
                keytypes.append(k.objtype)
        reqMsg = "{} {} {:n} {} {}".format(cmd,
                                             self.permutation.name,
                                             effectiveKeys,
                                             ' '.join(keynames),
                                             ' '.join(keytypes))
        repMsg = generic_msg(reqMsg)
        segAttr, uniqAttr = repMsg.split("+")
        if verbose: print(segAttr, uniqAttr)
        self.segments = create_pdarray(segAttr)
        unique_key_indices = create_pdarray(uniqAttr)
        if self.nkeys == 1:
            self.unique_keys = self.keys[unique_key_indices]
        else:
            self.unique_keys = [k[unique_key_indices] for k in self.keys]


    def count(self):
        '''
        Count the number of elements in each group, i.e. the number of times
        each key appears.

        Parameters
        ----------
        none

        Returns
        -------
        unique_keys : (list of) pdarray or Strings
            The unique keys, in grouped order
        counts : pdarray, int64
            The number of times each unique key appears
        
        '''
        if self.per_locale:
            cmd = "countLocalRdx"
        else:
            cmd = "countReduction"
        reqMsg = "{} {} {}".format(cmd, self.segments.name, self.size)
        repMsg = generic_msg(reqMsg)
        if verbose: print(repMsg)
        return self.unique_keys, create_pdarray(repMsg)
        
    def aggregate(self, values, operator):
        '''
        Using the permutation stored in the GroupBy instance, group another array 
        of values and apply a reduction to each group's values. 

        Parameters
        ----------
        values : pdarray
            The values to group and reduce
        operator: str
            The name of the reduction operator to use

        Returns
        -------
        unique_keys : (list of) pdarray or Strings
            The unique keys, in grouped order
        aggregates : pdarray
            One aggregate value per unique key in the GroupBy instance

        '''
        if not isinstance(values, pdarray):
            raise TypeError("<values> must be a pdarray")
        if values.size != self.size:
            raise ValueError("Attempt to group array using key array of different length")
        if operator not in self.Reductions:
            raise ValueError("Unsupported reduction: {}\nMust be one of {}".format(operator, self.Reductions))
        if self.assume_sorted:
            permuted_values = values
        else:
            permuted_values = values[self.permutation]
        if self.per_locale:
            cmd = "segmentedLocalRdx"
        else:
            cmd = "segmentedReduction"
        reqMsg = "{} {} {} {}".format(cmd,
                                         permuted_values.name,
                                         self.segments.name,
                                         operator)
        repMsg = generic_msg(reqMsg)
        if verbose: print(repMsg)
        if operator.startswith('arg'):
            return self.unique_keys, self.permutation[create_pdarray(repMsg)]
        else:
            return self.unique_keys, create_pdarray(repMsg)

    def sum(self, values):
        """
        Using the permutation stored in the GroupBy instance, group another array 
        of values and sum each group's values. 

        Parameters
        ----------
        values : pdarray
            The values to group and sum

        Returns
        -------
        unique_keys : (list of) pdarray or Strings
            The unique keys, in grouped order
        group_sums : pdarray
            One sum per unique key in the GroupBy instance

        Notes
        -----
        The grouped sum of a boolean ``pdarray`` returns integers.
        """
        return self.aggregate(values, "sum")
    
    def prod(self, values):
        """
        Using the permutation stored in the GroupBy instance, group another array 
        of values and compute the product of each group's values. 

        Parameters
        ----------
        values : pdarray
            The values to group and multiply

        Returns
        -------
        unique_keys : (list of) pdarray or Strings
            The unique keys, in grouped order
        group_products : pdarray, float64
            One product per unique key in the GroupBy instance

        Notes
        -----
        The return dtype is always float64.
        """
        return self.aggregate(values, "prod")
    
    def mean(self, values):
        """
        Using the permutation stored in the GroupBy instance, group another array 
        of values and compute the mean of each group's values. 

        Parameters
        ----------
        values : pdarray
            The values to group and average

        Returns
        -------
        unique_keys : (list of) pdarray or Strings
            The unique keys, in grouped order
        group_means : pdarray, float64
            One mean value per unique key in the GroupBy instance

        Notes
        -----
        The return dtype is always float64.
        """
        return self.aggregate(values, "mean")
    
    def min(self, values):
        """
        Using the permutation stored in the GroupBy instance, group another array 
        of values and return the minimum of each group's values. 

        Parameters
        ----------
        values : pdarray
            The values to group and find minima

        Returns
        -------
        unique_keys : (list of) pdarray or Strings
            The unique keys, in grouped order
        group_minima : pdarray
            One minimum per unique key in the GroupBy instance

        """
        return self.aggregate(values, "min")
    
    def max(self, values):
        """
        Using the permutation stored in the GroupBy instance, group another array 
        of values and return the maximum of each group's values. 

        Parameters
        ----------
        values : pdarray
            The values to group and find maxima

        Returns
        -------
        unique_keys : (list of) pdarray or Strings
            The unique keys, in grouped order
        group_maxima : pdarray
            One maximum per unique key in the GroupBy instance

        """
        return self.aggregate(values, "max")
    
    def argmin(self, values):
        """
        Using the permutation stored in the GroupBy instance, group another array 
        of values and return the location of the first minimum of each group's values. 

        Parameters
        ----------
        values : pdarray
            The values to group and find argmin

        Returns
        -------
        unique_keys : (list of) pdarray or Strings
            The unique keys, in grouped order
        group_argminima : pdarray, int64
            One index per unique key in the GroupBy instance

        Notes
        -----
        The returned indices refer to the original values array as passed in, not
        the permutation applied by the GroupBy instance.

        Examples
        --------
        >>> A = ak.array([0, 1, 0, 1, 0, 1])
        >>> B = ak.array([0, 1, 1, 0, 0, 1])
        >>> byA = ak.GroupBy(A)
        >>> byA.argmin(B)
        (array([0, 1]), array([0, 3]))
        """
        return self.aggregate(values, "argmin")
    
    def argmax(self, values):
        """
        Using the permutation stored in the GroupBy instance, group another array 
        of values and return the location of the first maximum of each group's values. 

        Parameters
        ----------
        values : pdarray
            The values to group and find argmax

        Returns
        -------
        unique_keys : (list of) pdarray or Strings
            The unique keys, in grouped order
        group_argmaxima : pdarray, int64
            One index per unique key in the GroupBy instance

        Notes
        -----
        The returned indices refer to the original values array as passed in, not
        the permutation applied by the GroupBy instance.

        Examples
        --------
        >>> A = ak.array([0, 1, 0, 1, 0, 1])
        >>> B = ak.array([0, 1, 1, 0, 0, 1])
        >>> byA = ak.GroupBy(A)
        >>> byA.argmax(B)
        (array([0, 1]), array([2, 1]))
        """
        return self.aggregate(values, "argmax")
    
    def nunique(self, values):
        """
        Using the permutation stored in the GroupBy instance, group another array 
        of values and return the number of unique values in each group. 

        Parameters
        ----------
        values : pdarray, int64
            The values to group and find unique values

        Returns
        -------
        unique_keys : (list of) pdarray or Strings
            The unique keys, in grouped order
        group_nunique : pdarray, int64
            Number of unique values per unique key in the GroupBy instance
        """
        return self.aggregate(values, "nunique")
    
    def any(self, values):
        """
        Using the permutation stored in the GroupBy instance, group another array 
        of values and perform an "or" reduction on each group. 

        Parameters
        ----------
        values : pdarray, bool
            The values to group and reduce with "or"

        Returns
        -------
        unique_keys : (list of) pdarray or Strings
            The unique keys, in grouped order
        group_any : pdarray, bool
            One bool per unique key in the GroupBy instance
        """
        return self.aggregate(values, "any")
    
    def all(self, values):
        """
        Using the permutation stored in the GroupBy instance, group another array 
        of values and perform an "and" reduction on each group. 

        Parameters
        ----------
        values : pdarray, bool
            The values to group and reduce with "and"

        Returns
        -------
        unique_keys : (list of) pdarray or Strings
            The unique keys, in grouped order
        group_any : pdarray, bool
            One bool per unique key in the GroupBy instance
        """
        return self.aggregate(values, "all")

    def broadcast(self, values):
        """
        Fill each group's segment with a constant value.

        Parameters
        ----------
        values : pdarray
            The values to put in each group's segment

        Returns
        -------
        pdarray
            The broadcast values

        Notes
        -----
        This function is a sparse analog of ``np.broadcast``. If a
        GroupBy object represents a sparse matrix (tensor), then
        this function takes a (dense) column vector and replicates
        each value to the non-zero elements in the corresponding row.

        The returned array is in permuted (grouped) order. To get
        back to the order of the array on which GroupBy was called,
        the user must invert the permutation (see below).

        Examples
        --------
        >>> a = ak.array([0, 1, 0, 1, 0])
        >>> values = ak.array([3, 5])
        >>> g = ak.GroupBy(a)
        # Result is in grouped order
        >>> g.broadcast(values)
        array([3, 3, 3, 5, 5]

        >>> b = ak.zeros_like(a)
        # Result is in original order
        >>> b[g.permutation] = g.broadcast(values)
        >>> b
        array([3, 5, 3, 5, 3])
        """

        if not isinstance(values, pdarray):
            raise ValueError("Vals must be pdarray")
        if values.size != self.segments.size:
            raise ValueError("Must have one value per segment")
        temp = zeros(self.size, values.dtype)
        if values.size == 0:
            return temp
        diffs = concatenate((array([values[0]]), values[1:] - values[:-1]))
        temp[self.segments] = diffs
        return cumsum(temp)



import numpy as np
import warnings
from itertools import product
from context import arkouda as ak
from base_test import ArkoudaTest
warnings.simplefilter("always", UserWarning)
SIZE = 10

'''
Tests the Arkouda where functionality and compares results to the
analogous Numpy were results.
'''
class WhereTest(ArkoudaTest):

    def test_where_equivalence(self):
        npA = {'int64': np.random.randint(0, 10, SIZE),
               'float64': np.random.randn(SIZE),
               'bool': np.random.randint(0, 2, SIZE, dtype='bool')}
        akA = {k: ak.array(v) for k, v in npA.items()}
        npB = {'int64': np.random.randint(10, 20, SIZE),
               'float64': np.random.randn(SIZE)+10,
               'bool': np.random.randint(0, 2, SIZE, dtype='bool')}
        akB = {k: ak.array(v) for k, v in npB.items()}
        npCond = np.random.randint(0, 2, SIZE, dtype='bool')
        akCond = ak.array(npCond)
        scA = {'int64': 42, 'float64': 2.71828, 'bool': True}
        scB = {'int64': -1, 'float64': 3.14159, 'bool': False}
        dtypes = set(npA.keys())
        failures = 0
        tests = 0
        for dtype in dtypes:
            for (ak1, ak2), (np1, np2) in zip(product((akA, scA), (akB, scB)),
                                              product((npA, scA), (npB, scB))):
                tests += 1
                akres = ak.where(akCond, ak1[dtype], ak2[dtype]).to_ndarray()
                npres = np.where(npCond, np1[dtype], np2[dtype])
                if not np.allclose(akres, npres, equal_nan=True):
                    self.assertWarning(warnings.warn("{} !=\n{}".format(akres, npres)))
                    failures += 1
        self.assertEqual(0,failures)



import numpy as np

from arkouda.client import generic_msg, verbose
from arkouda.dtypes import *
from arkouda.dtypes import structDtypeCodes, NUMBER_FORMAT_STRINGS
from arkouda.dtypes import dtype as akdtype
from arkouda.pdarrayclass import pdarray, create_pdarray
from arkouda.groupbyclass import GroupBy

global verbose

__all__ = ["join_on_eq_with_dt"]

predicates = {"true_dt":0, "abs_dt":1, "pos_dt":2}

def join_on_eq_with_dt(a1, a2, t1, t2, dt, pred, result_limit=1000):
    if not isinstance(a1, pdarray):
        raise ValueError("a1 must be pdarray")
    else:
        if not (a1.dtype == int64):
            raise ValueError("a1 must be int64 dtype")
        
    if not isinstance(a2, pdarray):
        raise ValueError("a2 must be pdarray")
    else:
        if not (a2.dtype == int64):
            raise ValueError("a2 must be int64 dtype")
        
    if not isinstance(t1, pdarray):
        raise ValueError("t1 must be pdarray")
    else:
        if not (t1.dtype == int64):
            raise ValueError("t1 must be int64 dtype")
        
    if not isinstance(t2, pdarray):
        raise ValueError("t2 must be pdarray")
    else:
        if not (t2.dtype == int64):
            raise ValueError("t2 must be int64 dtype")
        
    if not isinstance(dt, int):
        raise ValueError("dt must be an an int")
    
    if not (pred in predicates.keys()):
        raise ValueError("pred must be one of ",predicates.keys())
    
    if not isinstance(result_limit, int):
        raise ValueError("result_limit must be a scalar")

    # format numbers for request message
    dttype = resolve_scalar_dtype(dt)
    dtstr = NUMBER_FORMAT_STRINGS[dttype].format(dt)
    predtype = resolve_scalar_dtype(predicates[pred])
    predstr = NUMBER_FORMAT_STRINGS[predtype].format(predicates[pred])
    result_limittype = resolve_scalar_dtype(result_limit)
    result_limitstr = NUMBER_FORMAT_STRINGS[result_limittype].format(result_limit)
    # groupby on a2
    g2 = GroupBy(a2)
    # pass result into server joinEqWithDT operation
    repMsg = generic_msg("joinEqWithDT {} {} {} {} {} {} {} {} {}".format(a1.name,
                                                                          g2.segments.name,
                                                                          g2.unique_keys.name,
                                                                          g2.permutation.name,
                                                                          t1.name,
                                                                          t2.name,
                                                                          dtstr, predstr, result_limitstr))
    # create pdarrays for results
    resIAttr, resJAttr = repMsg.split("+")
    resI = create_pdarray(resIAttr)
    resJ = create_pdarray(resJAttr)
    return (resI, resJ)



    

    



from arkouda.client import generic_msg, verbose
from arkouda.pdarrayclass import pdarray, create_pdarray
from arkouda.pdarraycreation import zeros, array
from arkouda.sorting import argsort
from arkouda.strings import Strings

global verbose

__all__ = ["unique", "in1d", "concatenate", "union1d", "intersect1d",
           "setdiff1d", "setxor1d"]
global verbose

def unique(pda, return_counts=False):
    """
    Find the unique elements of an array.

    Returns the unique elements of an array, sorted if the values are integers. 
    There is an optional output in addition to the unique elements: the number 
    of times each unique value comes up in the input array.

    Parameters
    ----------
    pda : pdarray or Strings or Categorical
        Input array.
    return_counts : bool, optional
        If True, also return the number of times each unique item appears
        in `pda`.

    Returns
    -------
    unique : pdarray or Strings
        The unique values. If input dtype is int64, return values will be sorted.
    unique_counts : pdarray, optional
        The number of times each of the unique values comes up in the
        original array. Only provided if `return_counts` is True.

    Notes
    -----
    For integer arrays, this function checks to see whether `pda` is sorted and, if so,
    whether it is already unique. This step can save considerable computation.
    Otherwise, this function will sort `pda`. For 

    Examples
    --------
    >>> A = ak.array([3, 2, 1, 1, 2, 3])
    >>> ak.unique(A)
    array([1, 2, 3])
    """
    if hasattr(pda, 'unique'):
        return pda.unique()
    elif isinstance(pda, pdarray):
        repMsg = generic_msg("unique {} {} {}".format(pda.objtype, pda.name, return_counts))
        if return_counts:
            vc = repMsg.split("+")
            if verbose: print(vc)
            return create_pdarray(vc[0]), create_pdarray(vc[1])
        else:
            return create_pdarray(repMsg)
    elif isinstance(pda, Strings):
        name = '{}+{}'.format(pda.offsets.name, pda.bytes.name)
        repMsg = generic_msg("unique {} {} {}".format(pda.objtype, name, return_counts))
        vc = repMsg.split('+')
        if verbose: print(vc)
        if return_counts:
            return Strings(vc[0], vc[1]), create_pdarray(vc[2])
        else:
            return Strings(vc[0], vc[1])
    else:
        raise TypeError("must be pdarray or Strings {}".format(pda))

def in1d(pda1, pda2, invert=False):
    """
    Test whether each element of a 1-D array is also present in a second array.

    Returns a boolean array the same length as `pda1` that is True
    where an element of `pda1` is in `pda2` and False otherwise.

    Parameters
    ----------
    pda1 : pdarray or Strings or Categorical
        Input array.
    pda2 : pdarray or Strings
        The values against which to test each value of `pda1`. Must be the 
        same type as `pda1`.
    invert : bool, optional
        If True, the values in the returned array are inverted (that is,
        False where an element of `pda1` is in `pda2` and True otherwise).
        Default is False. ``ak.in1d(a, b, invert=True)`` is equivalent
        to (but is faster than) ``~ak.in1d(a, b)``.

    Returns
    -------
    pdarray, bool
        The values `pda1[in1d]` are in `pda2`.

    See Also
    --------
    unique, intersect1d, union1d

    Notes
    -----
    `in1d` can be considered as an element-wise function version of the
    python keyword `in`, for 1-D sequences. ``in1d(a, b)`` is logically
    equivalent to ``ak.array([item in b for item in a])``, but is much
    faster and scales to arbitrarily large ``a``.
    """
    if hasattr(pda1, 'in1d'):
        return pda1.in1d(pda2)
    elif isinstance(pda1, pdarray) and isinstance(pda2, pdarray):
        repMsg = generic_msg("in1d {} {} {}".format(pda1.name, pda2.name, invert))
        return create_pdarray(repMsg)
    elif isinstance(pda1, Strings) and isinstance(pda2, Strings):
        repMsg = generic_msg("segmentedIn1d {} {} {} {} {} {} {}".format(pda1.objtype,
                                                                         pda1.offsets.name,
                                                                         pda1.bytes.name,
                                                                         pda2.objtype,
                                                                         pda2.offsets.name,
                                                                         pda2.bytes.name,
                                                                         invert))
        return create_pdarray(repMsg)
    else:
        raise TypeError("must be pdarray {} or {}".format(pda1,pda2))

def concatenate(arrays):
    """
    Concatenate an iterable of ``pdarray`` objects into one ``pdarray``.

    Parameters
    ----------
    arrays : iterable of ``pdarray`` or Strings or Categorical
        The arrays to concatenate. Must all have same dtype.

    Returns
    -------
    pdarray
        Single array containing all values, in original order

    Examples
    --------
    >>> ak.concatenate([ak.array([1, 2, 3]), ak.array([4, 5, 6])])
    array([1, 2, 3, 4, 5, 6])
    """
    size = 0
    objtype = None
    dtype = None
    names = []
    if len(arrays) < 1:
        raise ValueError("concatenate called on empty iterable")
    if len(arrays) == 1:
        return arrays[0]
    if hasattr(arrays[0], 'concatenate'):
        return arrays[0].concatenate(arrays[1:])
    for a in arrays:
        if not isinstance(a, pdarray) and not isinstance(a, Strings):
            raise ValueError("Argument must be an iterable of pdarrays or Strings")
        if objtype == None:
            objtype = a.objtype
        if objtype == "pdarray":
            if dtype == None:
                dtype = a.dtype
            elif dtype != a.dtype:
                raise ValueError("All pdarrays must have same dtype")
            names.append(a.name)
        elif objtype == "str":
            names.append('{}+{}'.format(a.offsets.name, a.bytes.name))
        else:
            raise NotImplementedError("concatenate not implemented for object type {}".format(objtype))
        size += a.size
    if size == 0:
        if objtype == "pdarray":
            return zeros_like(arrays[0])
        else:
            return arrays[0]
    repMsg = generic_msg("concatenate {} {} {}".format(len(arrays), objtype, ' '.join(names)))
    if objtype == "pdarray":
        return create_pdarray(repMsg)
    elif objtype == "str":
        return Strings(*(repMsg.split('+')))

# (A1 | A2) Set Union: elements are in one or the other or both
def union1d(pda1, pda2):
    """
    Find the union of two arrays.

    Return the unique, sorted array of values that are in either of the two
    input arrays.

    Parameters
    ----------
    pda1 : pdarray
        Input array
    pda2 : pdarray
        Input array

    Returns
    -------
    pdarray
        Unique, sorted union of the input arrays.

    See Also
    --------
    intersect1d, unique

    Examples
    --------
    >>> ak.union1d([-1, 0, 1], [-2, 0, 2])
    array([-2, -1,  0,  1,  2])
    """
    if isinstance(pda1, pdarray) and isinstance(pda2, pdarray):
        if pda1.size == 0:
            return pda2 # union is pda2
        if pda2.size == 0:
            return pda1 # union is pda1
        if pda1.dtype == int and pda2.dtype == int:
            repMsg = generic_msg("union1d {} {}".format(pda1.name, pda2.name))
            return create_pdarray(repMsg)
        return unique(concatenate((unique(pda1), unique(pda2))))
    else:
        raise TypeError("must be pdarray {} or {}".format(pda1,pda2))

# (A1 & A2) Set Intersection: elements have to be in both arrays
def intersect1d(pda1, pda2, assume_unique=False):
    """
    Find the intersection of two arrays.

    Return the sorted, unique values that are in both of the input arrays.

    Parameters
    ----------
    pda1 : pdarray
        Input array
    pda2 : pdarray
        Input array
    assume_unique : bool
        If True, the input arrays are both assumed to be unique, which
        can speed up the calculation.  Default is False.

    Returns
    -------
    pdarray
        Sorted 1D array of common and unique elements.

    See Also
    --------
    unique, union1d

    Examples
    --------
    >>> ak.intersect1d([1, 3, 4, 3], [3, 1, 2, 1])
    array([1, 3])
    """
    if isinstance(pda1, pdarray) and isinstance(pda2, pdarray):
        if pda1.size == 0:
            return pda1 # nothing in the intersection
        if pda2.size == 0:
            return pda2 # nothing in the intersection
        if pda1.dtype == int and pda2.dtype == int:
            repMsg = generic_msg("intersect1d {} {} {}".format(pda1.name, pda2.name, assume_unique))
            return create_pdarray(repMsg)
        if not assume_unique:
            pda1 = unique(pda1)
            pda2 = unique(pda2)
        aux = concatenate((pda1, pda2))
        aux_sort_indices = argsort(aux)
        aux = aux[aux_sort_indices]
        mask = aux[1:] == aux[:-1]
        int1d = aux[:-1][mask]
        return int1d
    else:
        raise TypeError("must be pdarray {} or {}".format(pda1,pda2))

# (A1 - A2) Set Difference: elements have to be in first array but not second
def setdiff1d(pda1, pda2, assume_unique=False):
    """
    Find the set difference of two arrays.

    Return the sorted, unique values in `pda1` that are not in `pda2`.

    Parameters
    ----------
    pda1 : pdarray
        Input array.
    pda2 : pdarray
        Input comparison array.
    assume_unique : bool
        If True, the input arrays are both assumed to be unique, which
        can speed up the calculation.  Default is False.

    Returns
    -------
    pdarray
        Sorted 1D array of values in `pda1` that are not in `pda2`.

    See Also
    --------
    unique, setxor1d

    Examples
    --------
    >>> a = ak.array([1, 2, 3, 2, 4, 1])
    >>> b = ak.array([3, 4, 5, 6])
    >>> ak.setdiff1d(a, b)
    array([1, 2])
    """
    if isinstance(pda1, pdarray) and isinstance(pda2, pdarray):
        if pda1.size == 0:
            return pda1 # return a zero length pdarray
        if pda2.size == 0:
            return pda1 # subtracting nothing return orig pdarray
        if pda1.dtype == int and pda2.dtype == int:
            repMsg = generic_msg("setdiff1d {} {} {}".format(pda1.name, pda2.name, assume_unique))
            return create_pdarray(repMsg)
        if not assume_unique:
            pda1 = unique(pda1)
            pda2 = unique(pda2)
        return pda1[in1d(pda1, pda2, invert=True)]
    else:
        raise TypeError("must be pdarray {} or {}".format(pda1,pda2))

# (A1 ^ A2) Set Symmetric Difference: elements are not in the intersection
def setxor1d(pda1, pda2, assume_unique=False):
    """
    Find the set exclusive-or (symmetric difference) of two arrays.

    Return the sorted, unique values that are in only one (not both) of the
    input arrays.

    Parameters
    ----------
    pda1 : pdarray
        Input array.
    pda2 : pdarray
        Input array.
    assume_unique : bool
        If True, the input arrays are both assumed to be unique, which
        can speed up the calculation.  Default is False.

    Returns
    -------
    pdarray
        Sorted 1D array of unique values that are in only one of the input
        arrays.

    Examples
    --------
    >>> a = ak.array([1, 2, 3, 2, 4])
    >>> b = ak.array([2, 3, 5, 7, 5])
    >>> ak.setxor1d(a,b)
    array([1, 4, 5, 7])
    """
    if isinstance(pda1, pdarray) and isinstance(pda2, pdarray):
        if pda1.size == 0:
            return pda2 # return other pdarray if pda1 is empty
        if pda2.size == 0:
            return pda1 # return other pdarray if pda2 is empty
        if pda1.dtype == int and pda2.dtype == int:
            repMsg = generic_msg("setxor1d {} {} {}".format(pda1.name, pda2.name, assume_unique))
            return create_pdarray(repMsg)
        if not assume_unique:
            pda1 = unique(pda1)
            pda2 = unique(pda2)
        aux = concatenate((pda1, pda2))
        aux_sort_indices = argsort(aux)
        aux = aux[aux_sort_indices]
        flag = concatenate((array([True]), aux[1:] != aux[:-1], array([True])))
        return aux[flag[1:] & flag[:-1]]
    else:
        raise TypeError("must be pdarray {} or {}".format(pda1,pda2))



from arkouda.strings import Strings
from arkouda.pdarrayclass import pdarray
from arkouda.groupbyclass import GroupBy
from arkouda.pdarraycreation import zeros, zeros_like, arange
from arkouda.dtypes import int64, resolve_scalar_dtype
from arkouda.sorting import argsort
from arkouda.client import pdarrayIterThresh
from arkouda.pdarraysetops import unique, concatenate, in1d
import numpy as np

__all__ = ['Categorical']

class Categorical:
    """
    Represents an array of values belonging to named categories. Converting a Strings
    object to Categorical often saves memory and speeds up operations, especially
    if there are many repeated values, at the cost of some one-time work in initialization.

    Parameters
    ----------
    values : Strings
        String values to convert to categories 

    Attributes
    ----------
    categories : Strings
        The set of category labels (determined automatically)
    codes : pdarray, int64
        The category indices of the values or -1 for N/A
    permutation : pdarray, int64
        The permutation that groups the values in the same order as categories
    segments : pdarray, int64
        When values are grouped, the starting offset of each group
    size : int
        The number of items in the array
    nlevels : int
        The number of distinct categories
    ndim : int
        The rank of the array (currently only rank 1 arrays supported)
    shape : tuple
        The sizes of each dimension of the array
    """
    BinOps = frozenset(["==", "!="])
    objtype = "category"
    permutation = None
    segments = None
    
    def __init__(self, values, **kwargs):
        if 'codes' in kwargs and 'categories' in kwargs:
            # This initialization is called by Categorical.from_codes()
            # The values arg is ignored
            self.codes = kwargs['codes']
            self.categories = kwargs['categories']            
            if 'permutation' in kwargs:
                self.permutation = kwargs['permutation']
            if 'segments' in kwargs:
                self.segments = kwargs['segments']
        else:
            # Typical initialization, called with values
            if not isinstance(values, Strings):
                raise ValueError("Categorical: inputs other than Strings not yet supported")
            g = GroupBy(values)
            self.categories = g.unique_keys
            self.codes = zeros(values.size, dtype=int64)
            self.codes[g.permutation] = g.broadcast(arange(self.categories.size))
            self.permutation = g.permutation
            self.segments = g.segments
        # Always set these values
        self.size = self.codes.size
        self.nlevels = self.categories.size
        self.ndim = self.codes.ndim
        self.shape = self.codes.shape

    @classmethod
    def from_codes(cls, codes, categories, permutation=None, segments=None):
        """
        Make a Categorical from codes and categories arrays. If codes and categories
        have already been precomputed, this constructor saves time. If not, please
        use the normal constructor.

        Parameters
        ----------
        codes : pdarray, int64
            Category indices of each value
        categories : String
            Unique category labels
        """
        if not isinstance(codes, pdarray) or codes.dtype != int64:
            raise TypeError("Codes must be pdarray of int64")
        if not isinstance(categories, Strings):
            raise TypeError("Categories must be Strings")
        return cls(None, codes=codes, categories=categories, permutation=permutation, segments=segments)

    def to_ndarray(self):
        """
        Convert the array to a np.ndarray, transferring array data from the
        arkouda server to Python. This conversion discards category information
        and produces an ndarray of strings. If the array exceeds a builtin size limit,
        a RuntimeError is raised.

        Returns
        -------
        np.ndarray
            A numpy ndarray of strings corresponding to the values in this array

        Notes
        -----
        The number of bytes in the array cannot exceed ``arkouda.maxTransferBytes``,
        otherwise a ``RuntimeError`` will be raised. This is to protect the user
        from overflowing the memory of the system on which the Python client
        is running, under the assumption that the server is running on a
        distributed system with much more memory than the client. The user
        may override this limit by setting ak.maxTransferBytes to a larger
        value, but proceed with caution.
        """
        idx = self.categories.to_ndarray()
        valcodes = self.codes.to_ndarray()
        return idx[valcodes]

    def __iter__(self):
        return iter(self.to_ndarray())
        
    def __len__(self):
        return self.shape[0]

    def __str__(self):
        if self.size <= pdarrayIterThresh:
            vals = ["'{}'".format(self[i]) for i in range(self.size)]
        else:
            vals = ["'{}'".format(self[i]) for i in range(3)]
            vals.append('... ')
            vals.extend([self[i] for i in range(self.size-3, self.size)])
        return "[{}]".format(', '.join(vals))

    def __repr__(self):
        return "array({})".format(self.__str__())

    def binop(self, other, op):
        if op not in self.BinOps:
            raise NotImplementedError("Categorical: unsupported operator: {}".format(op))
        if np.isscalar(other) and resolve_scalar_dtype(other) == "str":
            idxresult = self.categories.binop(other, op)
            return idxresult[self.codes]
        if self.size != other.size:
            raise ValueError("Categorical {}: size mismatch {} {}".format(op, self.size, other.size))
        if isinstance(other, Categorical):
            if self.categories.name == other.categories.name:
                return self.codes.binop(other.codes, op)
            else:
                raise NotImplementedError("Operations between Categoricals with different indices not yet implemented")
        else:
            raise NotImplementedError("Operations between Categorical and non-Categorical not yet implemented. Consider converting operands to Categorical.")

    def r_binop(self, other, op):
        return self.binop(other, op)

    def __eq__(self, other):
        return self.binop(other, "==")

    def __neq__(self, other):
        return self.binop(other, "!=")

    def __getitem__(self, key):
        if np.isscalar(key) and resolve_scalar_dtype(key) == 'int64':
            return self.categories[self.codes[key]]
        else:
            return Categorical.from_codes(self.codes[key], self.categories)

    def reset_categories(self):
        """
        Recompute the category labels, discarding any unused labels. This method
        is often useful after slicing or indexing a Categorical array, when the
        resulting array only contains a subset of the original categories. In
        this case, eliminating unused categories can speed up other operations.
        """
        g = GroupBy(self.codes)
        idx = self.categories[g.unique_keys]
        newvals = zeros(self.codes.size, int64)
        newvals[g.permutation] = g.broadcast(arange(idx.size))
        return Categorical.from_codes(newvals, idx, permutation=g.permutation, segments=g.segments)

    def contains(self, substr):
        """
        Check whether each element contains the given substring.

        Parameters
        ----------
        substr : str
            The substring to search for

        Returns
        -------
        pdarray, bool
            True for elements that contain substr, False otherwise

        Notes
        -----
        This method can be significantly faster than the corresponding method
        on Strings objects, because it searches the unique category labels
        instead of the full array.

        See Also
        --------
        Categorical.startswith, Categorical.endswith
        """
        categoriescontains = self.categories.contains(substr)
        return categoriescontains[self.codes]

    def startswith(self, substr):
        """
        Check whether each element starts with the given substring.

        Parameters
        ----------
        substr : str
            The substring to search for

        Returns
        -------
        pdarray, bool
            True for elements that contain substr, False otherwise

        Notes
        -----
        This method can be significantly faster than the corresponding method
        on Strings objects, because it searches the unique category labels
        instead of the full array.

        See Also
        --------
        Categorical.contains, Categorical.endswith
        """
        categoriesstartswith = self.categories.startswith(substr)
        return categoriesstartswith[self.codes]

    def endswith(self, substr):
        """
        Check whether each element ends with the given substring.

        Parameters
        ----------
        substr : str
            The substring to search for

        Returns
        -------
        pdarray, bool
            True for elements that contain substr, False otherwise

        Notes
        -----
        This method can be significantly faster than the corresponding method
        on Strings objects, because it searches the unique category labels
        instead of the full array.

        See Also
        --------
        Categorical.startswith, Categorical.contains
        """
        categoriesendswith = self.categories.endswith(substr)
        return categoriesendswith[self.codes]

    def in1d(self, test):
        __doc__ = in1d.__doc__
        categoriesisin = in1d(self.categories, test)
        return categoriesisin[self.codes]

    def unique(self):
        __doc__ = unique.__doc__
        return Categorical.from_codes(arange(self.categories.size), self.categories)

    def group(self):
        """
        Return the permutation that groups the array, placing equivalent
        categories together. All instances of the same category are guaranteed to lie
        in one contiguous block of the permuted array, but the blocks are not
        necessarily ordered.

        Returns
        -------
        pdarray
            The permutation that groups the array by value

        See Also
        --------
        GroupBy, unique

        Notes
        -----
        This method is faster than the corresponding Strings method. If the Categorical
        was created from a Strings object, then this function simply returns the
        cached permutation. Even if the Categorical was created using from_codes(),
        this function will be faster than Strings.group() because it sorts dense
        integer values, rather than 128-bit hash values.
        """        
        if self.permutation is None:
            return argsort(self.codes)
        else:
            return self.permutation

    def argsort(self):
        __doc__ = argsort.__doc__
        idxperm = argsort(self.categories)
        inverse = zeros_like(idxperm)
        inverse[idxperm] = arange(idxperm.size)
        newvals = inverse[self.codes]
        return argsort(newvals)

    def sort(self):
        __doc__ = sort.__doc__
        idxperm = argsort(self.categories)
        inverse = zeros_like(idxperm)
        inverse[idxperm] = arange(idxperm.size)
        newvals = inverse[self.codes]
        return Categorical.from_codes(newvals, self.categories[idxperm])
            
    def merge(self, others):
        """
        Merge this Categorical with other Categoricals, concatenating the arrays and
        synchronizing the categories.

        Parameters
        ----------
        others : list of Categorical
            The Categorical arrays to concatenate and merge with this one

        Notes
        -----
        This operation can be expensive -- slower than concatenating Strings arrays.
        """
        if isinstance(others, Categorical):
            others = [others]
        elif len(others) < 1:
            return self
        samecategories = True
        for c in others:
            if not isinstance(c, Categorical):
                raise TypeError("Categorical: can only merge/concatenate with other Categoricals")
            if (self.categories.size != other.categories.size) or not (self.categories == other.categories).all():
                samecategories = False
        if samecategories:
            newvals = concatenate([self.codes] + [o.codes for o in others])
            return Categorical.from_codes(newvals, self.categories)
        else:
            g = ak.GroupBy(concatenate([self.categories] + [o.categories for o in others]))
            newidx = g.unique_keys
            wherediditgo = zeros(newidx.size, dtype=int64)
            wherediditgo[g.permutation] = arange(newidx.size)
            idxsizes = np.array([self.categories.size] + [o.categories.size for o in others])
            idxoffsets = np.cumsum(idxsizes) - idxsizes
            oldvals = concatenate([c.codes + off for c, off in zip([self.codes] + [o.codes for o in others], idxoffsets)])
            newvals = wherediditgo[oldvals]
            return Categorical.from_codes(newvals, newidx)
    #     msg = "segmentedMerge {} {} {} {} {} {}".format(self.categories.objtype,
    #                                                        self.categories.offsets.name,
    #                                                        self.categories.bytes.name,
    #                                                        other.categories.objtype,
    #                                                        other.categories.offsets.name,
    #                                                        other.categories.bytes.name)
    #     repMsg = generic_msg(msg)
    #     perm_attrib, seg_attrib, val_attrib = repMsg.split('+')
    #     perm = create_pdarray(perm_attrib)
    #     categories = Strings(seg_attrib, val_attrib)
    # #     return create_pdarray(repMsg)
        
    # # def merge(self, other):
    # #     perm = self.argmerge(other)
    # #     categories = concatenate((self.categories, other.categories))[perm]
    #     codes = zeros(self.size + other.size, int64)
    #     codes[:self.size] = perm[self.codes]
    #     codes[self.size:] = perm[other.codes + self.categories.size]
    #     return Categorical(codes, categories)



#!/usr/bin/env python3

import sys, time, argparse
import arkouda as ak

if __name__ == '__main__':
    ak.verbose = False
    parser = argparse.ArgumentParser()
    parser.add_argument('--server', default=None)
    parser.add_argument('--port', default=None)
    parser.add_argument('hdffiles', nargs='+')

    args = parser.parse_args()

    ak.set_defaults()
    ak.verbose = False
    if args.server is not None:
        if args.port is not None:
            ak.connect(server=args.server, port=args.port)
        else:
            ak.connect(server=args.server)
    else:
        if args.port is not None:
            ak.connect(port=args.port)
        else:
            ak.connect()

    print(ak.get_config())
            
    if len(args.hdffiles) == 0:
        print("usage: {} [--server server] [--port port] hdffiles ".format(sys.argv[0]))

    fields = ['srcIP', 'dstIP', 'srcPort', 'dstPort', 'start']
    
    nfDF = {field: ak.read_hdf(field, args.hdffiles) for field in fields}
    
    print(nfDF['start'])
    print(nfDF['srcIP'])
    print(nfDF['dstIP'])
    print(nfDF['srcPort'])
    print(nfDF['dstPort'])
    print(nfDF)
    
    print(ak.info(ak.AllSymbols))

    print("mem used: ", ak.get_mem_used())
    
    u,c = ak.unique(nfDF['srcIP'],return_counts=True)
    print("unique values = ", u.size,u)
    print("value counts = ", c.size,c)
    
    u,c = ak.unique(nfDF['dstIP'],return_counts=True)
    print("unique values = ", u.size,u)
    print("value counts = ", c.size,c)
    
    u,c = ak.unique(nfDF['srcPort'],return_counts=True)
    print("unique values = ", u.size,u)
    print("value counts = ", c.size,c)
    
    u,c = ak.unique(nfDF['dstPort'],return_counts=True)
    print("unique values = ", u.size,u)
    print("value counts = ", c.size,c)
    
    ak.shutdown()
    #ak.disconnect()
    



from arkouda.client import *
from arkouda.dtypes import *
from arkouda.pdarrayclass import *
from arkouda.sorting import *
from arkouda.pdarraysetops import * 
from arkouda.pdarraycreation import *
from arkouda.numeric import *
from arkouda.pdarrayIO import *
from arkouda.groupbyclass import *
from arkouda.strings import *
from arkouda.join import *
from arkouda.registration import *
from arkouda.categorical import *



import numpy as np
import numpy
import warnings, os
from itertools import product
from base_test import ArkoudaTest
from context import arkouda as ak
from context import arkouda
SIZE = 10
verbose = ArkoudaTest.verbose

def run_tests(verbose):
    # ignore numpy warnings like divide by 0
    np.seterr(all='ignore')
    global pdarrays
    pdarrays = {'int64': ak.arange(0, SIZE, 1),
                'float64': ak.linspace(0, 2, SIZE),
                'bool': (ak.arange(0, SIZE, 1) % 2) == 0}
    global ndarrays
    ndarrays = {'int64': np.arange(0, SIZE, 1),
                'float64': np.linspace(0, 2, SIZE),
                'bool': (np.arange(0, SIZE, 1) % 2) == 0}
    global scalars
    #scalars = {k: v[SIZE//2] for k, v in ndarrays.items()}
    scalars = {'int64': 5,
               'float64': 3.14159,
               'bool': True}
    dtypes = pdarrays.keys()
    if verbose:
        print("Operators: ", ak.pdarray.BinOps)
        print("Dtypes: ", dtypes)
        print("pdarrays: ")
        for k, v in pdarrays.items():
            print(k, ": ", v)
        print("ndarrays: ")
        for k, v in ndarrays.items():
            print(k, ": ", v)
        print("scalars: ")
        for k, v in scalars.items():
            print(k, ": ", v)

    def do_op(lt, rt, ls, rs, isarkouda, oper):
        evalstr = ''
        if ls:
            evalstr += 'scalars["{}"]'.format(lt)
        else:
            evalstr += '{}["{}"]'.format(('ndarrays', 'pdarrays')[isarkouda], lt)
        evalstr += ' {} '.format(oper)
        if rs:
            evalstr += 'scalars["{}"]'.format(rt)
        else:
            evalstr += '{}["{}"]'.format(('ndarrays', 'pdarrays')[isarkouda], rt)
        #print(evalstr)
        res = eval(evalstr)
        return res

    results = {'neither_implement': [],   # (expression, ak_error)
               'arkouda_minus_numpy': [], # (expression, ak_result, error_on_exec?)
               'numpy_minus_arkouda': [], # (expression, ak_result, error_on_exec?)
               'both_implement': []}      # (expression, ak_result, error_on_exec?, dtype_mismatch?, value_mismatch?)
    tests = 0
    for ltype, rtype, op in product(dtypes, dtypes, ak.pdarray.BinOps):
        for lscalar, rscalar in ((False, False), (False, True), (True, False)):
            tests += 1
            expression = "{}({}) {} {}({})".format(ltype, ('array', 'scalar')[lscalar], op, rtype, ('array', 'scalar')[rscalar])
            try:
                npres = do_op(ltype, rtype, lscalar, rscalar, False, op)
            except TypeError: # numpy doesn't implement operation
                try:
                    akres = do_op(ltype, rtype, lscalar, rscalar, True, op)
                except RuntimeError as e:
                    if 'not implemented' in str(e): # neither numpy nor arkouda implement
                        results['neither_implement'].append((expression, str(e)))
                    else: # arkouda implements with error, np does not implement
                        results['arkouda_minus_numpy'].append((expression, str(e), True))
                    continue
                # arkouda implements but not numpy
                results['arkouda_minus_numpy'].append((expression, str(akres), False))
                continue
            try:
                akres = do_op(ltype, rtype, lscalar, rscalar, True, op)
            except RuntimeError as e:
                if 'not implemented' in str(e): # numpy implements but not arkouda
                    results['numpy_minus_arkouda'].append((expression, str(e), True))
                else: # both implement, but arkouda errors
                    results['both_implement'].append((expression, str(e), True, False, False))
                continue
            # both numpy and arkouda execute without error
            try:
                akrestype = akres.dtype
            except Exception as e:
                warnings.warn("Cannot detect return dtype of ak result: {} (np result: {})".format(akres, npres))
                results['both_implement'].append((expression, str(akres), False, True, False))
                continue

            if akrestype != npres.dtype:
                restypes = "{}(np) vs. {}(ak)".format(npres.dtype, akrestype)
                #warnings.warn("dtype mismatch: {} = {}".format(expression, restypes))
                results['both_implement'].append((expression, restypes, False, True, False))
                continue
            try:
                akasnp = akres.to_ndarray()
            except Exception as e:
                warnings.warn("Could not convert to ndarray: {}".format(akres))
                results['both_implement'].append((expression, str(akres), True, False, False))
                continue
            if not np.allclose(akasnp, npres, equal_nan=True):
                res = "np: {}\nak: {}".format(npres, akasnp)
                # warnings.warn("result mismatch: {} =\n{}".format(expression, res))
                results['both_implement'].append((expression, res, False, False, True))
            # Finally, both numpy and arkouda agree on result
            results['both_implement'].append((expression, "", False, False, False))

    print("# ops not implemented by numpy or arkouda: {}".format(len(results['neither_implement'])))
    if verbose:
        for expression, err in results['neither_implement']:
            print(expression)
    print("# ops implemented by numpy but not arkouda: {}".format(len(results['numpy_minus_arkouda'])))
    if verbose:
        for expression, err, flag in results['numpy_minus_arkouda']:
            print(expression)
    print("# ops implemented by arkouda but not numpy: {}".format(len(results['arkouda_minus_numpy'])))
    if verbose:
        for expression, res, flag in results['arkouda_minus_numpy']:
            print(expression, " -> ", res)
    nboth = len(results['both_implement'])
    print("# ops implemented by both: {}".format(nboth))
    matches = 0
    execerrors = []
    dtypeerrors = []
    valueerrors = []
    for (expression, res, ex, dt, val) in results['both_implement']:
        matches += not any((ex, dt, val))
        if ex: execerrors.append((expression, res))
        if dt: dtypeerrors.append((expression, res))
        if val: valueerrors.append((expression, res))
    print("  Matching results:         {} / {}".format(matches, nboth))
    print("  Arkouda execution errors: {} / {}".format(len(execerrors), nboth))
    if verbose: print('\n'.join(map(': '.join, execerrors)))
    print("  Dtype mismatches:         {} / {}".format(len(dtypeerrors), nboth))
    if verbose: print('\n'.join(map(': '.join, dtypeerrors)))
    print("  Value mismatches:         {} / {}".format(len(valueerrors), nboth))
    if verbose: print('\n'.join(map(': '.join, valueerrors)))
    return matches == nboth

'''
Encapsulates test cases that invoke the run_tests method.
'''
class OperatorsTest(ArkoudaTest):

    def testPdArrayAddInt(self):
        aArray = ak.ones(100)
        addArray = aArray + 1
        self.assertIsInstance(addArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(2),addArray[0])

        addArray = 1 + aArray
        self.assertIsInstance(addArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(2), addArray[0])

    def testPdArrayAddNumpyInt(self):
        aArray = ak.ones(100)
        addArray = aArray + np.int64(1)
        self.assertIsInstance(addArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(2), addArray[0])

        addArray = np.int64(1) + aArray
        self.assertIsInstance(addArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(2), addArray[0])

    def testPdArraySubtractInt(self):
        aArray = ak.ones(100)
        subArray =  aArray - 2
        self.assertIsInstance(subArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(-1), subArray[0])

        subArray =  2 - aArray
        self.assertIsInstance(subArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(1), subArray[0])

    def testPdArraySubtractNumpyInt(self):
        aArray = ak.ones(100)
        subArray =  aArray - np.int64(2)
        self.assertIsInstance(subArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(-1), subArray[0])

        subArray =  np.int64(2) - aArray
        self.assertIsInstance(subArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(1), subArray[0])

    def testPdArrayMultInt(self):
        aArray = ak.ones(100)
        mArray =  aArray*5
        self.assertIsInstance(mArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(5), mArray[0])

        mArray =  5*aArray
        self.assertIsInstance(mArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(5), mArray[0])

    def testPdArrayMultNumpyInt(self):
        aArray = ak.ones(100)
        mArray =  aArray*np.int64(5)
        self.assertIsInstance(mArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(5), mArray[0])

        mArray =  np.int64(5)*aArray
        self.assertIsInstance(mArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(5), mArray[0])

    def testPdArrayDivideInt(self):
        aArray = ak.ones(100)
        dArray =  aArray*15/3
        self.assertIsInstance(dArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(5), dArray[0])

        dArray =  15*aArray/3
        self.assertIsInstance(dArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(5), dArray[0])

    def testPdArrayDivideNumpyInt(self):
        aArray = ak.ones(100)
        dArray =  aArray*np.int64(15)/3
        self.assertIsInstance(dArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(5), dArray[0])

        dArray =  np.int64(15)*aArray/3
        self.assertIsInstance(dArray, ak.pdarrayclass.pdarray)
        self.assertEqual(np.float64(5), dArray[0])
        
    def testAllOperators(self):
        run_tests(verbose)
        
if __name__ == '__main__':
    '''
    Enables invocation of operator tests outside of pytest test harness
    '''
    import sys
    if len(sys.argv) not in (3, 4):
        print(f"Usage: {sys.argv[0]} <server_name> <port> [<verbose>=(0|1)]")
    verbose = False
    if len(sys.argv) == 4 and sys.argv[3] == "1":
        verbose = True
    ak.connect(server=sys.argv[1], port=int(sys.argv[2]))
    success = run_tests(verbose)
    ak.disconnect()
    sys.exit((1, 0)[success])



import numpy as np

from arkouda.client import generic_msg
from arkouda.dtypes import *
from arkouda.pdarrayclass import pdarray, parse_single_value, create_pdarray
from arkouda.pdarraysetops import unique

__all__ = ["abs", "log", "exp", "cumsum", "cumprod", "sin", "cos", "where",
           "histogram", "value_counts"]

def abs(pda):
    """
    Return the element-wise absolute value of the array.
    """
    if isinstance(pda, pdarray):
        repMsg = generic_msg("efunc {} {}".format("abs", pda.name))
        return create_pdarray(repMsg)
    else:
        raise TypeError("must be pdarray {}".format(pda))

def log(pda):
    """
    Return the element-wise natural log of the array. 

    Notes
    -----
    Logarithms with other bases can be computed as follows:

    >>> A = ak.array([1, 10, 100])
    # Natural log
    >>> ak.log(A)
    array([0, 2.3025850929940459, 4.6051701859880918])
    # Log base 10
    >>> ak.log(A) / np.log(10)
    array([0, 1, 2])
    # Log base 2
    >>> ak.log(A) / np.log(2)
    array([0, 3.3219280948873626, 6.6438561897747253])
    """
    if isinstance(pda, pdarray):
        repMsg = generic_msg("efunc {} {}".format("log", pda.name))
        return create_pdarray(repMsg)
    else:
        raise TypeError("must be pdarray {}".format(pda))

def exp(pda):
    """
    Return the element-wise exponential of the array.
    """
    if isinstance(pda, pdarray):
        repMsg = generic_msg("efunc {} {}".format("exp", pda.name))
        return create_pdarray(repMsg)
    else:
        raise TypeError("must be pdarray {}".format(pda))

def cumsum(pda):
    """
    Return the cumulative sum over the array. 

    The sum is inclusive, such that the ``i`` th element of the 
    result is the sum of elements up to and including ``i``.
    """
    if isinstance(pda, pdarray):
        repMsg = generic_msg("efunc {} {}".format("cumsum", pda.name))
        return create_pdarray(repMsg)
    else:
        raise TypeError("must be pdarray {}".format(pda))

def cumprod(pda):
    """
    Return the cumulative product over the array. 

    The product is inclusive, such that the ``i`` th element of the 
    result is the product of elements up to and including ``i``.
    """
    if isinstance(pda, pdarray):
        repMsg = generic_msg("efunc {} {}".format("cumprod", pda.name))
        return create_pdarray(repMsg)
    else:
        raise TypeError("must be pdarray {}".format(pda))

def sin(pda):
    """
    Return the element-wise sine of the array.
    """
    if isinstance(pda,pdarray):
        repMsg = generic_msg("efunc {} {}".format("sin",pda.name))
        return create_pdarray(repMsg)
    else:
        raise TypeError("must be pdarray {}".format(pda))

def cos(pda):
    """
    Return the element-wise cosine of the array.
    """
    if isinstance(pda,pdarray):
        repMsg = generic_msg("efunc {} {}".format("cos",pda.name))
        return create_pdarray(repMsg)
    else:
        raise TypeError("must be pdarray {}".format(pda))
    
def where(condition, A, B):
    """
    Return an array with elements chosen from A and B based on a conditioning array.
    
    Parameters
    ----------
    condition : pdarray
        Used to choose values from A or B
    A : scalar or pdarray
        Value(s) used when condition is True
    B : scalar or pdarray
        Value(s) used when condition is False

    Returns
    -------
    pdarray
        Values chosen from A and B according to condition

    Notes
    -----
    A and B must have the same dtype.
    """
    if not isinstance(condition, pdarray):
        raise TypeError("must be pdarray {}".format(condition))
    if isinstance(A, pdarray) and isinstance(B, pdarray):
        repMsg = generic_msg("efunc3vv {} {} {} {}".format("where",
                                                           condition.name,
                                                           A.name,
                                                           B.name))
    # For scalars, try to convert it to the array's dtype
    elif isinstance(A, pdarray) and np.isscalar(B):
        repMsg = generic_msg("efunc3vs {} {} {} {} {}".format("where",
                                                              condition.name,
                                                              A.name,
                                                              A.dtype.name,
                                                              A.format_other(B)))
    elif isinstance(B, pdarray) and np.isscalar(A):
        repMsg = generic_msg("efunc3sv {} {} {} {} {}".format("where",
                                                              condition.name,
                                                              B.dtype.name,
                                                              B.format_other(A),
                                                              B.name))
    elif np.isscalar(A) and np.isscalar(B):
        # Scalars must share a common dtype (or be cast)
        dtA = resolve_scalar_dtype(A)
        dtB = resolve_scalar_dtype(B)
        # Make sure at least one of the dtypes is supported
        if not (dtA in DTypes or dtB in DTypes):
            raise TypeError("Not implemented for scalar types {} and {}".format(dtA, dtB))
        # If the dtypes are the same, do not cast
        if dtA == dtB:
            dt = dtA
        # If the dtypes are different, try casting one direction then the other
        elif dtB in DTypes and np.can_cast(A, dtB):
            A = np.dtype(dtB).type(A)
            dt = dtB
        elif dtA in DTypes and np.can_cast(B, dtA):
            B = np.dtype(dtA).type(B)
            dt = dtA
        # Cannot safely cast
        else:
            raise TypeError("Cannot cast between scalars {} and {} to supported dtype".format(A, B))
        repMsg = generic_msg("efunc3ss {} {} {} {} {} {}".format("where",
                                                                 condition.name,
                                                                 dt,
                                                                 A,
                                                                 dt,
                                                                 B))
    return create_pdarray(repMsg)


def histogram(pda, bins=10):
    """
    Compute a histogram of evenly spaced bins over the range of an array.
    
    Parameters
    ----------
    pda : pdarray
        The values to histogram

    bins : int
        The number of equal-size bins to use (default: 10)

    Returns
    -------
    pdarray
        The number of values present in each bin

    See Also
    --------
    value_counts

    Notes
    -----
    The bins are evenly spaced in the interval [pda.min(), pda.max()]. Currently,
    the user must re-compute the bin edges, e.g. with np.linspace (see below) 
    in order to plot the histogram.

    Examples
    --------
    >>> A = ak.arange(0, 10, 1)
    >>> nbins = 3
    >>> h = ak.histogram(A, bins=nbins)
    >>> h
    array([3, 3, 4])
    # Recreate the bin edges in NumPy
    >>> binEdges = np.linspace(A.min(), A.max(), nbins+1)
    >>> binEdges
    array([0., 3., 6., 9.])
    # To plot, use only the left edges, and export the histogram to NumPy
    >>> plt.plot(binEdges[:-1], h.to_ndarray())
    """
    if isinstance(pda, pdarray) and isinstance(bins, int):
        repMsg = generic_msg("histogram {} {}".format(pda.name, bins))
        return create_pdarray(repMsg)
    else:
        raise TypeError("must be pdarray {} and bins must be an int {}".format(pda,bins))


def value_counts(pda):
    """
    Count the occurrences of the unique values of an array.

    Parameters
    ----------
    pda : pdarray, int64
        The array of values to count

    Returns
    -------
    unique_values : pdarray, int64
        The unique values, sorted in ascending order

    counts : pdarray, int64
        The number of times the corresponding unique value occurs

    See Also
    --------
    unique, histogram

    Notes
    -----
    This function differs from ``histogram()`` in that it only returns counts 
    for values that are present, leaving out empty "bins".

    Examples
    --------
    >>> A = ak.array([2, 0, 2, 4, 0, 0])
    >>> ak.value_counts(A)
    (array([0, 2, 4]), array([3, 2, 1]))
    """
    return unique(pda, return_counts=True)



import numpy as np
from context import arkouda as ak
from arkouda import dtypes
from base_test import ArkoudaTest

'''
DtypesTest encapsulates arkouda dtypes module methods
'''
class DtypesTest(ArkoudaTest):

    def test_check_np_dtype(self):
      
        '''
        Tests dtypes.check_np_dtype method 
        
        :return: None
        :raise: AssertionError if 1.. test cases fail
        '''
        dtypes.check_np_dtype(np.dtype(np.bool))
        dtypes.check_np_dtype(np.dtype(np.int64))
        dtypes.check_np_dtype(np.dtype(np.float64))
        dtypes.check_np_dtype(np.dtype(np.uint8))
        dtypes.check_np_dtype(np.dtype(np.str))
        with self.assertRaises(TypeError):
            dtypes.check_np_dtype(np.dtype(np.int16))

    def test_translate_np_dtype(self):
        '''
        Tests dtypes.translate_np_dtype method
        
        :return: None
        :raise: AssertionError if 1.. test cases fail
        '''
        d_tuple = dtypes.translate_np_dtype(np.dtype(np.bool))
        self.assertEqual(1, d_tuple[1])
        self.assertEqual('bool', d_tuple[0])
        d_tuple = dtypes.translate_np_dtype(np.dtype(np.int64))
        self.assertEqual(8, d_tuple[1])
        self.assertEqual('int', d_tuple[0])  
        d_tuple = dtypes.translate_np_dtype(np.dtype(np.float64))
        self.assertEqual(8, d_tuple[1])
        self.assertEqual('float', d_tuple[0])        
        d_tuple = dtypes.translate_np_dtype(np.dtype(np.uint8))
        self.assertEqual(1, d_tuple[1])
        self.assertEqual('uint',d_tuple[0])    
        d_tuple = dtypes.translate_np_dtype(np.dtype(np.str))
        self.assertEqual(0, d_tuple[1])
        self.assertEqual('str', d_tuple[0])   
        with self.assertRaises(TypeError):
            dtypes.check_np_dtype(np.dtype(np.int16))
            
    def test_resolve_scalar_dtype(self):
        '''
        Tests dtypes.resolve_scalar_dtype method
        
        :return: None
        :raise: AssertionError if 1.. test cases fail
        '''
        self.assertEqual('bool', dtypes.resolve_scalar_dtype(True))
        self.assertEqual('int64', dtypes.resolve_scalar_dtype(1))
        self.assertEqual('float64', dtypes.resolve_scalar_dtype(float(0.0)))
        self.assertEqual('str', dtypes.resolve_scalar_dtype('test'))
        self.assertEqual('int64', dtypes.resolve_scalar_dtype(np.int64(1))) 
        self.assertEqual("<class 'list'>", dtypes.resolve_scalar_dtype([1]))
      


import numpy as np
import struct

from arkouda.client import generic_msg, maxTransferBytes
from arkouda.dtypes import *
from arkouda.dtypes import structDtypeCodes, NUMBER_FORMAT_STRINGS
from arkouda.dtypes import dtype as akdtype
from arkouda.pdarrayclass import pdarray, create_pdarray
from arkouda.strings import Strings

__all__ = ["array", "zeros", "ones", "zeros_like", "ones_like", "arange",
           "linspace", "randint", "uniform", "standard_normal",
           "random_strings_uniform", "random_strings_lognormal"]

def array(a):
    """
    Convert an iterable to a pdarray, sending data to the arkouda server.

    Parameters
    ----------
    a : array_like
        Rank-1 array of a supported dtype

    Returns
    -------
    pdarray
        Instance of pdarray stored on arkouda server

    See Also
    --------
    pdarray.to_ndarray

    Notes
    -----
    The number of bytes in the input array cannot exceed `arkouda.maxTransferBytes`,
    otherwise a RuntimeError will be raised. This is to protect the user
    from overwhelming the connection between the Python client and the arkouda
    server, under the assumption that it is a low-bandwidth connection. The user
    may override this limit by setting ak.maxTransferBytes to a larger value, 
    but should proceed with caution.

    Examples
    --------
    >>> a = [3, 5, 7]
    >>> b = ak.array(a)
    >>> b
    array([3, 5, 7])
   
    >>> type(b)
    arkouda.pdarray    
    """
    # If a is already a pdarray, do nothing
    if isinstance(a, pdarray):
        return a
    # If a is not already a numpy.ndarray, convert it
    if not isinstance(a, np.ndarray):
        try:
            a = np.array(a)
        except:
            raise TypeError("Argument must be array-like")
    # Only rank 1 arrays currently supported
    if a.ndim != 1:
        raise RuntimeError("Only rank-1 arrays supported")
    # Check if array of strings
    if a.dtype.kind == 'U':
        # Length of each string, plus null byte terminator
        lengths = np.array([len(elem) for elem in a]) + 1
        # Compute zero-up segment offsets
        offsets = np.cumsum(lengths) - lengths
        # Allocate and fill bytes array with string segments
        nbytes = offsets[-1] + lengths[-1]
        if nbytes > maxTransferBytes:
            raise RuntimeError("Creating pdarray would require transferring {} bytes, which exceeds allowed transfer size. Increase ak.maxTransferBytes to force.".format(nbytes))
        values = np.zeros(nbytes, dtype=np.uint8)
        for s, o in zip(a, offsets):
            for i, b in enumerate(s.encode()):
                values[o+i] = b
        # Recurse to create pdarrays for offsets and values, then return Strings object
        return Strings(array(offsets), array(values))
    # If not strings, then check that dtype is supported in arkouda
    if a.dtype.name not in DTypes:
        raise RuntimeError("Unhandled dtype {}".format(a.dtype))
    # Do not allow arrays that are too large
    size = a.size
    if (size * a.itemsize) > maxTransferBytes:
        raise RuntimeError("Array exceeds allowed transfer size. Increase ak.maxTransferBytes to allow")
    # Pack binary array data into a bytes object with a command header
    # including the dtype and size
    fmt = ">{:n}{}".format(size, structDtypeCodes[a.dtype.name])
    req_msg = "array {} {:n} ".format(a.dtype.name, size).encode() + struct.pack(fmt, *a)
    rep_msg = generic_msg(req_msg, send_bytes=True)
    return create_pdarray(rep_msg)

def zeros(size, dtype=np.float64):
    """
    Create a pdarray filled with zeros.

    Parameters
    ----------
    size : int
        Size of the array (only rank-1 arrays supported)
    dtype : {float64, int64, bool}
        Type of resulting array, default float64

    Returns
    -------
    pdarray
        Zeros of the requested size and dtype

    See Also
    --------
    ones, zeros_like

    Examples
    --------
    >>> ak.zeros(5, dtype=ak.int64)
    array([0, 0, 0, 0, 0])
    >>> ak.zeros(5, dtype=ak.float64)
    array([0, 0, 0, 0, 0])
    >>> ak.zeros(5, dtype=ak.bool)
    array([False, False, False, False, False])
    """
    if not np.isscalar(size):
        raise TypeError("size must be a scalar, not {}".format(type(size)))
    dtype = akdtype(dtype) # normalize dtype
    # check dtype for error
    if dtype.name not in DTypes:
        raise TypeError("unsupported dtype {}".format(dtype))
    kind, itemsize = translate_np_dtype(dtype)
    repMsg = generic_msg("create {} {}".format(dtype.name, size))
    return create_pdarray(repMsg)

def ones(size, dtype=float64):
    """
    Create a pdarray filled with ones.

    Parameters
    ----------
    size : int
        Size of the array (only rank-1 arrays supported)
    dtype : {float64, int64, bool}
        Resulting array type, default float64

    Returns
    -------
    pdarray
        Ones of the requested size and dtype

    See Also
    --------
    zeros, ones_like

    Examples
    --------
    >>> ak.ones(5, dtype=ak.int64)
    array([1, 1, 1, 1, 1])
    >>> ak.ones(5, dtype=ak.float64)
    array([1, 1, 1, 1, 1])
    >>> ak.ones(5, dtype=ak.bool)
    array([True, True, True, True, True])
    """
    if not np.isscalar(size):
        raise TypeError("size must be a scalar, not {}".format(type(size)))
    dtype = akdtype(dtype) # normalize dtype
    # check dtype for error
    if dtype.name not in DTypes:
        raise TypeError("unsupported dtype {}".format(dtype))
    kind, itemsize = translate_np_dtype(dtype)
    repMsg = generic_msg("create {} {}".format(dtype.name, size))
    a = create_pdarray(repMsg)
    a.fill(1)
    return a

def zeros_like(pda):
    """
    Create a zero-filled pdarray of the same size and dtype as an existing pdarray.

    Parameters
    ----------
    pda : pdarray
        Array to use for size and dtype

    Returns
    -------
    pdarray
        Equivalent to ak.zeros(pda.size, pda.dtype)

    See Also
    --------
    zeros, ones_like
    """
    if isinstance(pda, pdarray):
        return zeros(pda.size, pda.dtype)
    else:
        raise TypeError("must be pdarray {}".format(pda))

def ones_like(pda):
    """
    Create a one-filled pdarray of the same size and dtype as an existing pdarray.

    Parameters
    ----------
    pda : pdarray
        Array to use for size and dtype

    Returns
    -------
    pdarray
        Equivalent to ak.ones(pda.size, pda.dtype)

    See Also
    --------
    ones, zeros_like
    """
    if isinstance(pda, pdarray):
        return ones(pda.size, pda.dtype)
    else:
        raise TypeError("must be pdarray {}".format(pda))

def arange(*args):
    """
    arange([start,] stop[, stride])

    Create a pdarray of consecutive integers within the interval [start, stop).
    If only one arg is given then arg is the stop parameter. If two args are given
    then the first arg is start and second is stop. If three args are given
    then the first arg is start, second is stop, third is stride.

    Parameters
    ----------
    start : int, optional
        Starting value (inclusive), the default starting value is 0
    stop : int
        Stopping value (exclusive)
    stride : int, optional
        The difference between consecutive elements, the default stride is 1,
        if stride is specified then start must also be specified

    Returns
    -------
    pdarray, int64
        Integers from start (inclusive) to stop (exclusive) by stride

    See Also
    --------
    linspace, zeros, ones, randint
    
    Notes
    -----
    Negative strides result in decreasing values. Currently, only int64 pdarrays
    can be created with this function. For float64 arrays, use linspace.

    Examples
    --------
    >>> ak.arange(0, 5, 1)
    array([0, 1, 2, 3, 4])

    >>> ak.arange(5, 0, -1)
    array([5, 4, 3, 2, 1])

    >>> ak.arange(0, 10, 2)
    array([0, 2, 4, 6, 8])
    """
   
    #if one arg is given then arg is stop
    if len(args) == 1:
        start = 0
        stop = args[0]
        stride = 1

    #if two args are given then first arg is start and second is stop
    if len(args) == 2:
        start = args[0]
        stop = args[1]
        stride = 1

    #if three args are given then first arg is start,
    #second is stop, third is stride
    if len(args) == 3:
        start = args[0]
        stop = args[1]
        stride = args[2]

    if not all((np.isscalar(start), np.isscalar(stop), np.isscalar(stride))):
        raise TypeError("all arguments must be scalars")

    if stride == 0:
        raise ZeroDivisionError("division by zero")

    if isinstance(start, int) and isinstance(stop, int) and isinstance(stride, int):
        # TO DO: fix bug in server that goes 2 steps too far for negative strides
        if stride < 0:
            stop = stop + 2
        repMsg = generic_msg("arange {} {} {}".format(start, stop, stride))
        return create_pdarray(repMsg)
    else:
        raise TypeError("start,stop,stride must be type int {} {} {}".format(start,stop,stride))

def linspace(start, stop, length):
    """
    Create a pdarray of linearly spaced points in a closed interval.

    Parameters
    ----------
    start : scalar
        Start of interval (inclusive)
    stop : scalar
        End of interval (inclusive)
    length : int
        Number of points

    Returns
    -------
    pdarray, float64
        Array of evenly spaced points along the interval

    See Also
    --------
    arange

    Examples
    --------
    >>> ak.linspace(0, 1, 5)
    array([0, 0.25, 0.5, 0.75, 1])
    """
    if not all((np.isscalar(start), np.isscalar(stop), np.isscalar(length))):
        raise TypeError("all arguments must be scalars")
    starttype = resolve_scalar_dtype(start)
    startstr = NUMBER_FORMAT_STRINGS[starttype].format(start)
    stoptype = resolve_scalar_dtype(stop)
    stopstr = NUMBER_FORMAT_STRINGS[stoptype].format(stop)
    lentype = resolve_scalar_dtype(length)
    if lentype != 'int64':
        raise TypeError("Length must be int64")
    lenstr = NUMBER_FORMAT_STRINGS[lentype].format(length)
    repMsg = generic_msg("linspace {} {} {}".format(startstr, stopstr, lenstr))
    return create_pdarray(repMsg)


def randint(low, high, size, dtype=int64):
    """
    Generate a pdarray with random values in a specified range.

    Parameters
    ----------
    low : int
        The low value (inclusive) of the range
    high : int
        The high value (exclusive for int, inclusive for float) of the range
    size : int
        The length of the returned array
    dtype : {int64, float64, bool}
        The dtype of the array

    Returns
    -------
    pdarray
        Values drawn uniformly from the specified range having the desired dtype

    Notes
    -----
    Calling randint with dtype=float64 will result in uniform non-integral
    floating point values.

    Examples
    --------
    >>> ak.randint(0, 10, 5)
    array([5, 7, 4, 8, 3])

    >>> ak.randint(0, 1, 3, dtype=ak.float64)
    array([0.92176432277231968, 0.083130710959903542, 0.68894208386667544])

    >>> ak.randint(0, 1, 5, dtype=ak.bool)
    array([True, False, True, True, True])
    """
    if not all((np.isscalar(low), np.isscalar(high), np.isscalar(size))):
        raise TypeError("all arguments must be scalars")
    if resolve_scalar_dtype(size) != 'int64':
        raise TypeError("size must be integer")
    if size < 0 or high < low:
        raise ValueError("Incompatible arguments")
    dtype = akdtype(dtype) # normalize dtype
    # check dtype for error
    if dtype.name not in DTypes:
        raise TypeError("unsupported dtype {}".format(dtype))
    lowstr = NUMBER_FORMAT_STRINGS[dtype.name].format(low)
    highstr = NUMBER_FORMAT_STRINGS[dtype.name].format(high)
    sizestr = NUMBER_FORMAT_STRINGS['int64'].format(size)
    repMsg = generic_msg("randint {} {} {} {}".format(sizestr, dtype.name, lowstr, highstr))
    return create_pdarray(repMsg)


def uniform(size, low=0.0, high=1.0):
    """
    Generate a pdarray with uniformly distributed random values in a specified range.

    Parameters
    ----------
    low : int
        The low value (inclusive) of the range
    high : int
        The high value (inclusive) of the range
    size : int
        The length of the returned array

    Returns
    -------
    pdarray, float64
        Values drawn uniformly from the specified range

    Examples
    --------
    >>> ak.uniform(3)
    array([0.92176432277231968, 0.083130710959903542, 0.68894208386667544])
    """
    return randint(size, low=low, high=high, dtype='float64')
    

def standard_normal(size):
    """
    Draw real numbers from the standard normal distribution.

    Parameters
    ----------
    size : int
        The number of samples to draw (size of the returned array)
    
    Returns
    -------
    pdarray, float64
        The array of random numbers

    See Also
    --------
    randint

    Notes
    -----
    For random samples from :math:`N(\mu, \sigma^2)`, use:

    ``(sigma * standard_normal(size)) + mu``
    """
    if size < 0:
        raise ValueError("Invalid size: {}".format(size))
    msg = "randomNormal {}".format(NUMBER_FORMAT_STRINGS['int64'].format(size))
    repMsg = generic_msg(msg)
    return create_pdarray(repMsg)


def random_strings_uniform(minlen, maxlen, size, characters='uppercase'):
    """
    Generate random strings with lengths uniformly distributed between 
    minlen and maxlen, and with characters drawn from a specified set.

    Parameters
    ----------
    minlen : int
        The minimum allowed length of string
    maxlen : int
        The maximum allowed length of string
    size : int
        The number of strings to generate
    characters : (uppercase, lowercase, numeric, printable, binary)
        The set of characters to draw from

    Returns
    -------
    Strings
        The array of random strings

    See Also
    --------
    random_strings_lognormal, randint
    """
    if minlen < 0 or maxlen < minlen or size < 0:
        raise ValueError("Incompatible arguments")
    msg = "randomStrings {} {} {} {} {}".format(NUMBER_FORMAT_STRINGS['int64'].format(size),
                                                "uniform",
                                                characters,
                                                NUMBER_FORMAT_STRINGS['int64'].format(minlen),
                                                NUMBER_FORMAT_STRINGS['int64'].format(maxlen))
    repMsg = generic_msg(msg)
    return Strings(*(repMsg.split('+')))


def random_strings_lognormal(logmean, logstd, size, characters='uppercase'):
    """
    Generate random strings with log-normally distributed lengths and 
    with characters drawn from a specified set.

    Parameters
    ----------
    logmean : float
        The log-mean of the length distribution
    logstd : float
        The log-standard-deviation of the length distribution
    size : int
        The number of strings to generate
    characters : (uppercase, lowercase, numeric, printable, binary)
        The set of characters to draw from

    Returns
    -------
    Strings
        The array of random strings

    See Also
    --------
    random_strings_lognormal, randint

    Notes
    -----
    The lengths of the generated strings are distributed $Lognormal(\mu, \sigma^2)$,
    with :math:`\mu = logmean` and :math:`\sigma = logstd`. Thus, the strings will have
    an average length of :math:`exp(\mu + 0.5*\sigma^2)`, a minimum length of zero, and
    a heavy tail towards longer strings.
    """
    if logstd <= 0 or size < 0:
        raise ValueError("Incompatible arguments")
    msg = "randomStrings {} {} {} {} {}".format(NUMBER_FORMAT_STRINGS['int64'].format(size),
                                                "lognormal",
                                                characters,
                                                NUMBER_FORMAT_STRINGS['float64'].format(logmean),
                                                NUMBER_FORMAT_STRINGS['float64'].format(logstd))
    repMsg = generic_msg(msg)
    return Strings(*(repMsg.split('+')))



import json, os

from arkouda.client import generic_msg
from arkouda.pdarrayclass import pdarray, create_pdarray
from arkouda.strings import Strings

__all__ = ["ls_hdf", "read_hdf", "read_all", "load", "get_datasets",
           "load_all", "save_all"]

def ls_hdf(filename):
    """
    This function calls the h5ls utility on a filename visible to the arkouda
    server.

    Parameters
    ----------
    filename : str
        The name of the file to pass to h5ls

    Returns
    -------
    str
        The string output of `h5ls <filename>` from the server
    """
    return generic_msg("lshdf {}".format(json.dumps([filename])))

def read_hdf(dsetName, filenames):
    """
    Read a single dataset from multiple HDF5 files into an arkouda pdarray.

    Parameters
    ----------
    dsetName : str
        The name of the dataset (must be the same across all files)
    filenames : list or str
        Either a list of filenames or shell expression

    Returns
    -------
    pdarray
        A pdarray instance pointing to the server-side data read in

    See Also
    --------
    get_datasets, ls_hdf, read_all, load, save

    Notes
    -----
    If filenames is a string, it is interpreted as a shell expression
    (a single filename is a valid expression, so it will work) and is
    expanded with glob to read all matching files. Use ``get_datasets`` to
    show the names of datasets in HDF5 files.

    If dsetName is not present in all files, a RuntimeError is raised.
    """
    if isinstance(filenames, str):
        filenames = [filenames]
    if not isinstance(dsetName, str):
        raise TypeError("Datset name must be a str, not {}".format(type(dsetName)))
    # rep_msg = generic_msg("readhdf {} {:n} {}".format(dsetName, len(filenames), json.dumps(filenames)))
    # # This is a hack to detect a string return type
    # # In the future, we should put the number and type into the return message
    # if '+' in rep_msg:
    #     return Strings(*rep_msg.split('+'))
    # else:
    #     return create_pdarray(rep_msg)
    return read_all(filenames, datasets=dsetName)

def read_all(filenames, datasets=None, iterative=False):
    """
    Read datasets from HDF5 files.

    Parameters
    ----------
    filenames : list or str
        Either a list of filenames or shell expression
    datasets : list or str or None
        (List of) name(s) of dataset(s) to read (default: all available)
    iterative : boolean
        Iterative (True) or Single (False) function call(s) to server

    Returns
    -------
    For a single dataset returns an Arkouda pdarray or an Arkouda Sring and
    for multiple datasets returns a dictionary of Ardkouda pdarrays and
    Arkouda Strings.
        Dictionary of {datasetName: pdarray or String}

    See Also
    --------
    read_hdf, get_datasets, ls_hdf

    Notes
    -----
    If filenames is a string, it is interpreted as a shell expression
    (a single filename is a valid expression, so it will work) and is
    expanded with glob to read all matching files.

    If iterative == True each dataset name and file names are passed to
    the server as independent sequential strings while if iterative == False
    all dataset names and file names are passed to the server in a single
    string.

    If datasets is None, infer the names of datasets from the first file
    and read all of them. Use ``get_datasets`` to show the names of datasets in
    HDF5 files.

    If not all datasets are present in all HDF5 files, a RuntimeError
    is raised.a

    """
    if isinstance(filenames, str):
        filenames = [filenames]
    if datasets is None:
        datasets = get_datasets(filenames[0])
    if isinstance(datasets, str):
        datasets = [datasets]
    else: # ensure dataset(s) exist
        if isinstance(datasets, str):
            datasets = [datasets]
        nonexistent = set(datasets) - set(get_datasets(filenames[0]))
        if len(nonexistent) > 0:
            raise ValueError("Dataset(s) not found: {}".format(nonexistent))
    if iterative == True: # iterative calls to server readhdf
        return {dset:read_hdf(dset, filenames) for dset in datasets}
    else:  # single call to server readAllHdf
        rep_msg = generic_msg("readAllHdf {:n} {:n} {} | {}".format(len(datasets), len(filenames), json.dumps(datasets), json.dumps(filenames)))
        if ',' in rep_msg:
            rep_msgs = rep_msg.split(' , ')
            d = dict()
            for dset, rm in zip(datasets, rep_msgs):
                    if('+' in rm): #String
                        d[dset]=Strings(*rm.split('+'))
                    else:
                        d[dset]=create_pdarray(rm)
            return d
        elif '+' in rep_msg:
            return Strings(*rep_msg.split('+'))
        else:
            return create_pdarray(rep_msg)

def load(path_prefix, dataset='array'):
    """

    Load a pdarray previously saved with ``pdarray.save()``.

    Parameters
    ----------
    path_prefix : str
        Filename prefix used to save the original pdarray
    dataset : str
        Dataset name where the pdarray was saved

    Returns
    -------
    pdarray
        The pdarray that was previously saved

    See Also
    --------
    save, load_all, read_hdf, read_all
    """
    prefix, extension = os.path.splitext(path_prefix)
    globstr = "{}_LOCALE*{}".format(prefix, extension)
    return read_hdf(dataset, globstr)

def get_datasets(filename):
    """
    Get the names of datasets in an HDF5 file.

    Parameters
    ----------
    filename : str
        Name of an HDF5 file visible to the arkouda server

    Returns
    -------
    list of str
        Names of the datasets in the file

    See Also
    --------
    ls_hdf
    """
    rep_msg = ls_hdf(filename)
    datasets = [line.split()[0] for line in rep_msg.splitlines()]
    return datasets

def load_all(path_prefix):
    """
    Load multiple pdarray previously saved with ``save_all()``.

    Parameters
    ----------
    path_prefix : str
        Filename prefix used to save the original pdarray

    Returns
    -------
    dict of pdarrays
        Dictionary of {datsetName: pdarray} with the previously saved pdarrays

    See Also
    --------
    save_all, load, read_hdf, read_all
    """
    prefix, extension = os.path.splitext(path_prefix)
    firstname = "{}_LOCALE0{}".format(prefix, extension)
    return {dataset: load(path_prefix, dataset=dataset) for dataset in get_datasets(firstname)}

def save_all(columns, path_prefix, names=None, mode='truncate'):
    """
    Save multiple named pdarrays to HDF5 files.

    Parameters
    ----------
    columns : dict or list of pdarrays
        Collection of arrays to save
    path_prefix : str
        Directory and filename prefix for output files
    names : list of str
        Dataset names for the pdarrays
    mode : {'truncate' | 'append'}
        By default, truncate (overwrite) the output files if they exist.
        If 'append', attempt to create new dataset in existing files.

    See Also
    --------
    save, load_all

    Notes
    -----
    Creates one file per locale containing that locale's chunk of each pdarray.
    If columns is a dictionary, the keys are used as the HDF5 dataset names.
    Otherwise, if no names are supplied, 0-up integers are used. By default,
    any existing files at path_prefix will be overwritten, unless the user
    specifies the 'append' mode, in which case arkouda will attempt to add
    <columns> as new datasets to existing files. If the wrong number of files
    is present or dataset names already exist, a RuntimeError is raised.
    """
    if names is not None and len(names) != len(columns):
        raise ValueError("Number of names does not match number of columns")
    if isinstance(columns, dict):
        pdarrays = columns.values()
        if names is None:
            names = columns.keys()
    elif isinstance(columns, list):
        pdarrays = columns
        if names is None:
            names = range(len(columns))
    if (mode.lower() not in 'append') and (mode.lower() not in 'truncate'):
        raise ValueError("Allowed modes are 'truncate' and 'append'")
    first_iter = True
    for arr, name in zip(pdarrays, names):
        # Append all pdarrays to existing files as new datasets EXCEPT the first one, and only if user requests truncation
        if mode.lower() not in 'append' and first_iter:
            arr.save(path_prefix, dataset=name, mode='truncate')
            first_iter = False
        else:
            arr.save(path_prefix, dataset=name, mode='append')



from context import arkouda as ak
import numpy as np
from time import time
from base_test import ArkoudaTest
OPERATORS = ['sum', 'min', 'nunique']

def generate_arrays(length, nkeys, nvals, dtype='int64'):
    keys = ak.randint(0, nkeys, length)
    if dtype == 'int64':
        vals = ak.randint(0, nvals, length)
    elif dtype == 'bool':
        vals = ak.zeros(length, dtype='bool')
        for i in np.random.randint(0, length, nkeys//2):
            vals[i] = True
    else:
        vals = ak.linspace(-1, 1, length)
    return keys, vals

if __name__ == '__main__':
    import sys
    if len(sys.argv) != 7:
        print(f"Usage: {sys.argv[0]} <server> <port> <strategy (0=global, 1=perLocale)> <length> <num_keys> <num_vals>")
        sys.exit()
    per_locale = (sys.argv[3] == '1')
    print("per_locale = ", per_locale)
    length = int(sys.argv[4])
    print("length     = ", length)
    nkeys = int(sys.argv[5])
    print("nkeys      = ", nkeys)
    nvals = int(sys.argv[6])
    print("nvals      = ", nvals)
    ak.connect(sys.argv[1], int(sys.argv[2]))
    print("Generating keys and vals...")
    start = time()
    keys, vals = generate_arrays(length, nkeys, nvals)
    print(f"{time() - start:.2f} seconds", end="\n\n")
    print("GroupBy...")
    start = time()
    g = ak.GroupBy(keys, per_locale)
    print(f"{time() - start:.2f} seconds", end="\n\n")
    for op in OPERATORS:
        print(f"Aggregate('{op}') ...")
        start = time()
        uk, rv = g.aggregate(vals, op)
        print(f"{time() - start:.2f} seconds", end="\n\n")
    sys.exit()



# !/usr/bin/env python3
# test data located at  https://csr.lanl.gov/data/netflow.html
# field names located at https://csr.lanl.gov/data/2017.html

from context import arkouda as ak
import sys, os, h5py, time
from glob import glob
from base_test import ArkoudaTest


class ReadAllTest(ArkoudaTest):

    def setUp(self):
      
        '''
        Invokes the parent setUp method to start the arkouda_server and also
        sets the test_data_url used to lookup test files.sets
        
        :return: None
        :raise: AssertionError if the TEST_DATA_URL has not been set
        '''
        ArkoudaTest.setUp(self)
        self.test_data_url = os.getenv('TEST_DATA_URL')
        assert self.test_data_url, 'The TEST_DATA_URL env variable must be set'

    def testReadAll(self):
        ak.verbose = False    #client verbose Flag
        cwd = os.getcwd()
        allfiles = glob(cwd+'/../converter/netflow_day-*.hdf')
        print(allfiles)
        start = time.time()
        dictionary1 = ak.read_all(allfiles, iterative=True)
        end = time.time()
        t1 = end - start
        print("read_all(iterative=True) seconds: %.3f" % (t1))
        for key, value in dictionary1.items():
            print(key,type(value),value,len(value))

        start = time.time()
        dictionary2 = ak.read_all(allfiles)
        end = time.time()
        t2 = end - start
        print("read_all() seconds: %.3f" % (t2))
        for key, value in dictionary2.items():
            print(key,type(value),value,len(value))
        
if __name__ == '__main__':        
    if len(sys.argv) < 3:
        print("Usage: {} <hostname> <port> <HDF5_filenames>".format(sys.argv[0]))
        sys.exit()
    ak.connect(sys.argv[1], sys.argv[2])
    ak.verbose = False    #client verbose Flag
    cwd = os.getcwd()
    allfiles = glob(cwd+'/../converter/netflow_day-*.hdf')
    if len(sys.argv) > 3:
        allfiles = sys.argv[3:]

    start = time.time()
    dictionary1 = ak.read_all(allfiles, iterative=True)
    end = time.time()
    t1 = end - start
    print("read_all(iterative=True) seconds: %.3f" % (t1))
    for key, value in dictionary1.items():
        print(key,type(value),value,len(value))

    start = time.time()
    dictionary2 = ak.read_all(allfiles)
    end = time.time()
    t2 = end - start
    print("read_all() seconds: %.3f" % (t2))
    for key, value in dictionary2.items():
        print(key,type(value),value,len(value))

    ak.disconnect()
